class nioDict (dict):
    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def __getitem__(self, key):
        #print 'in nioDict getitem', self.path, repr(self.topdict)
        if key[0] == '/' and len(key) > 1:
            val = dict.__getitem__(self, key[1:])
        elif key.find('/') > 0 and not self.topdict is None:
            newkey = ''.join([self.path,'/',key])
            val = dict.__getitem__(self.topdict, newkey)
        else:
            val = dict.__getitem__(self, key)
#        print 'GET', key
        return val

    def __setitem__(self, key, val):
#        print 'SET', key, val
        dict.__setitem__(self, key, val)

    def __repr__(self):
        dictrepr = dict.__repr__(self)
        #return '%s(%s)' % (type(self).__name__, dictrepr)
        return  dictrepr

    def update(self, *args, **kwargs):
        #print 'update', args, kwargs
        for k, v in dict(*args, **kwargs).iteritems():
            self[k] = v
   




version = '1.5.0'
array_module = 'numpy'
array_module_version = '1.11.3'
python_version = '2.7.1'
formats = {}
formats['hdfeos5'] = 0
formats['netcdf4'] = 1
formats['hdf4'] = 1
formats['hdf5'] = 1
formats['shapefile'] = 1
formats['grib2'] = 1
formats['hdfeos'] = 1
formats['opendap'] = 1



#$Id$

'''
PyNIO enables NetCDF-like access for multiple file formats including
    NetCDF 3 (rw) and  GRIB 1 (r), and optionally
    NetCDF 4 (rw), HDF 4 (rw), HDF 5 (rw), HDFEOS 2/4 (r), GRIB 2 (r), HDFEOS 5 (r), shapefiles
    and other GDAL OGR-supported (r) data formats.

import Nio

Class NioFile:

f = Nio.open_file(filepath, mode='r', options=None, history='',format='')

To see summary of file contents, including all attributes :
   print f
attributes:
   name -- the name of the file or group
   dimensions -- dictionary of dimension lengths with dimension name keys
   variables -- dictionary of variable objects with variable name keys
   attributes --  dictionary of global file or group attributes with attribute name keys
       (the following are applicable for advanced formats NetCDF4 and HDF5 only)
   groups -- dictionary of groups with group name keys
   ud_types -- dictionary of user-defined data type definitions with data type name keys
   chunk_dimensions -- dictionary of chunking dimension sizes with dimension name keys
   parent -- reference to the parent group, parent file for the root group, or None for a file
   path -- the path of a group relative to the root group ('/'), or the file path for a file
methods:
   close(history='')
   create_dimension(name, length)
   create_variable(name, type, dimensions)
   unlimited(dimension_name) -- returns True if dimension_name refers to an unlimited dimension; False otherwise
       (the following are applicable for advanced formats NetCDF4 and HDF5 only)
   create_group(name) -- create a group in the file or group.
   create_vlen(name,type,dimensions) -- create a variable length array variable in the file or group.
   create_compound(name,type,dimensions) -- create a compound variable with the given type and dimensions.
   create_compound_type(name, type)  -- create a user-defined compound type; with member names, sizes
        and types as defined in the type sequence argument.
   
For more detailed information:
    print f.__doc__

Class NioOptions

opt = Nio.options()

To set general or format-specific options assign option names and settings as attributes
and values of 'opt'. Then pass 'opt' as the optional options argument to
Nio.open_file.
To see valid options:
    print opt.__doc__ 

Class NioVariable

v = f.variables['varname']

To see summary of variable contents including all dimensions,
associated coordinate variables, and attributes:
    print v 
Attributes:
    rank -- a scalar value indicating the number of dimensions
    shape -- a tuple containing the number of elements in each dimension
    dimensions -- a tuple containing the dimensions names in order
    attributes -- a dictionary of variable attributes with attribute name keys
    size -- a scalar value indicating the size in bytes of the variable
    name -- the name of the variable
    parent -- reference to the group or file to which the variable belongs
    path -- the path of the variable relative to the root group ('/')
Methods:
    assign_value(value) -- assign a value to a variable in the file.
    get_value() -- retrieve the value of a variable in the file.
    typecode() -- return a character code representing the variable's type.
    set_option(option,value) -- set certain options.

For more detailed information:
    print v.__doc__

For complete documentation see:

        http://www.pyngl.ucar.edu/Nio.html
'''

from nio import *
from nio import _C_API

#
#  Get version number and flag for NumPy compatibility.
#
#  Also, get the __array_module__  and __array_module_version__
#  attributes.
#
import pynio_version
__version__              = pynio_version.version
__array_module__         = pynio_version.array_module
__array_module_version__ = pynio_version.array_module_version
__formats__              = pynio_version.formats
del pynio_version

NioVariable = None
NioFile = None
__all__ = [ 'open_file', 'option_defaults', 'options' ]

def pyniopath_ncarg():
#
#  Find the root directory that contains the supplemental PyNIO files,
#  in particular, the grib2 codetables. For now the default is to look
#  for a path in sys.path that contains the subdirectory PyNIO/ncarg. 
#   Otherwise, check NCARG_ROOT/lib/ncarg.
#
    import sys
    pynio_ncarg = None
    for path in sys.path:
        trypath = os.path.join(path,"PyNIO","ncarg")
        if not os.path.exists(trypath):
            trypath = os.path.join(path,"ncarg")
            if not os.path.exists(os.path.join(trypath,"grib2_codetables")):
                continue
        if os.path.exists(trypath):
            pynio_ncarg = trypath
            return pynio_ncarg

    if pynio_ncarg == None:
        ncarg_dir = os.environ.get("NCARG_ROOT")
        if ncarg_dir == None or not os.path.exists(ncarg_dir) \
          or not os.path.exists(os.path.join(ncarg_dir,"lib","ncarg")):
            if not __formats__['grib2']:
                return ""
            else:
                print "No path found to PyNIO/ncarg data directory and no usable NCARG installation found"
                sys.exit()
        else:
            pynio_ncarg = os.path.join(ncarg_dir,"lib","ncarg")

    print pynio_ncarg
    return pynio_ncarg

#
# Set the NCARG_NCARG environment variable.
# This should allow the grib2_codetables directory to be found without
# requiring any environment variables to be set by the user
# Also set NIO_GRIB2_CODETABLES if it is not set already

import os
os.environ["NCARG_NCARG"] = pyniopath_ncarg()
if not os.getenv("NIO_GRIB2_CODETABLES"):
    os.environ["NIO_GRIB2_CODETABLES"] = os.path.join(os.environ["NCARG_NCARG"],"grib2_codetables") 
del pyniopath_ncarg

#
# This part of the code creates the NioFile and NioVariable classes as
# proxies for the C API private classes _NioFile and _NioVariable
# It also implements support for Juerg Schmidli's coordinate selection 
# code


from nio import _Nio
import numpy as np
from numpy import ma
from coordsel import get_variable, inp2csel, inp2isel, inp2xsel, idxsel2xsel, \
                     xSelect, crdSelect, idxSelect 
import coordsel as cs
import niodict  as nd

_Nio.option_defaults['UseAxisAttribute'] = False
_Nio.option_defaults['MaskedArrayMode'] = 'MaskedIfFillAtt'
_Nio.option_defaults['ExplicitFillValues'] = None
_Nio.option_defaults['MaskBelowValue'] = None
_Nio.option_defaults['MaskAboveValue'] = None

def get_integer_version(strversion):
    ''' converts string version number into an integer '''
    d = strversion.split('.')
    if len(d) > 2:
       v = int(d[0]) * 10000 + int(d[1]) * 100 + int(d[2][0])
    elif len(d) is 2:
       v = int(d[0]) * 10000 + int(d[1]) * 100
    else:
       v = int(d[0]) * 10000
    return v

_is_new_ma = get_integer_version(np.__version__) > 10004
del get_integer_version

_builtins = ['__class__', '__delattr__', '__doc__', '__getattribute__', '__hash__', \
            '__init__', '__module__', '__new__', '__reduce__', '__reduce_ex__', '__repr__', \
            '__setattr__', '__str__', '__weakref__', '__getitem__', '__setitem__', '__len__' ]

_localatts = ['open', 'attributes','_obj','variables','file','varname', 'groups', 'parent', \
              'cf_dimensions', 'cf2dims', 'ma_mode', 'explicit_fill_values', 'coordinates', \
              'mask_below_value', 'mask_above_value', 'set_option', 'create_variable', 'create_group', 'close', 'assign_value', 'get_value']    

# The Proxy concept is adapted from Martelli, Ravenscroft, & Ascher, 'Python Cookbook' Second Edition, 6.6
import weakref

class _Proxy(object):
    ''' base class for all proxies '''


    def __init__(self, obj):
        super(_Proxy, self).__init__()
        super(_Proxy,self).__setattr__('_obj', obj)
        super(_Proxy,self).__setattr__('attributes',{})
        for key in obj.__dict__.keys():
            if key != 'attributes':
                super(_Proxy,self).__setattr__(key,obj.__dict__[key])
            self.attributes[key] = obj.__dict__[key]

    def __getattribute__(self, attrib):
        if attrib in _localatts or attrib in _builtins:
            return super(_Proxy,self).__getattribute__(attrib)
        else:
            return getattr(self._obj,attrib)

    def __setattr__(self, attrib, value):
        if attrib in _builtins:
            raise AttributeError, "Attempt to modify read only attribute"
        elif attrib in _localatts:
            super(_Proxy,self).__setattr__(attrib,value)
        else:
            setattr(self._obj,attrib,value)
            self.attributes[attrib] = self._obj.__dict__[attrib]

    def __delattr__(self, attrib):
        if attrib in _localatts or attrib in _builtins:
            raise AttributeError, "Attempt to delete read only attribute"
        else:
            delattr(self._obj,attrib)
            del(self.attributes[attrib])

def _make_binder(unbound_method):
    def f(self, *a, **k): return unbound_method(self._obj, *a, **k)
    # in 2.4, only: f.__name__ = unbound_method.__name__
    return f

_known_proxy_classes = {  }

def _proxy(obj, *specials, **regulars):
    ''' factory-function for a proxy able to delegate special methods '''
    # do we already have a suitable customized class around?
    obj_cls = type(obj)
    key = obj_cls, specials
    cls = _known_proxy_classes.get(key)
    if cls is None:
        # we don't have a suitable class around, so let's make it
        cls_dict = {}
        cls_dict['__doc__'] = obj_cls.__doc__
        # this removes the underscore in the name as supplied by the C API module
        # to give the user-visible name
        name = obj_cls.__name__[1:] 
        cls = type(name, (_Proxy,), cls_dict)
        for name in specials:
            name = '__%s__' % name
            unbound_method = getattr(obj_cls, name)
            setattr(cls, name, _make_binder(unbound_method))
        # also cache it for the future
        _known_proxy_classes[key] = cls
        for key in regulars.keys():
            setattr(cls, key, regulars[key])
    # instantiate and return the needed proxy
    instance = cls(obj)
    return instance



def _fill_value_to_masked(self, a):

    #
    # _FillValue is the preferred fill value attribute but if it is not set
    # then look for missing_value
    #
    if self.file.ma_mode == 'maskednever':
        # MaskedNever -- just return a numpy array
         return a
    elif self.file.ma_mode == 'maskedexplicit':
        # handle user-specified masking -- first ranges, then explicit single values
        # note that masked_where does not remove previously applied mask values 
        # so it can be applied in stages

        if self.file.mask_below_value is not None and self.file.mask_above_value is not None:
            if self.file.mask_below_value > self.file.mask_above_value: 
                # mask a band of values
                a = ma.masked_where((a < self.file.mask_below_value) & (a > self.file.mask_above_value),a,copy=0)
            else:
                # mask high and low values                    
                a = ma.masked_where((a < self.file.mask_below_value) | (a > self.file.mask_above_value),a,copy=0)
        elif self.file.mask_below_value is not None:
            # mask low values
            a = ma.masked_where(a < self.file.mask_below_value,a,copy=0)
        elif self.file.mask_above_value is not None:
            # mask high values
            a = ma.masked_where(a > self.file.mask_above_value,a,copy=0)

        # now apply single fill values
        if hasattr(self.file.explicit_fill_values,'__iter__'):
            # multiple explicit fill values
            for fval in self.file.explicit_fill_values:
                a = ma.masked_where(a == fval,a,copy=0)
            a.set_fill_value(self.file.explicit_fill_values[0])
        elif self.file.explicit_fill_values is not None:
            a = ma.masked_where(a == self.file.explicit_fill_values,a,copy=0)
            a.set_fill_value(self.file.explicit_fill_values)

    elif self.file.ma_mode == 'maskediffillattandvalue':
        # MaskedIfFillAttAndValue -- return a masked array only if there are actual fill values
        if self.__dict__.has_key('_FillValue'):
            if a.__contains__(self.__dict__['_FillValue']):
                if np.isscalar(a):
                    a = ma.masked_where(a == self.__dict__['_FillValue'][0],a,copy=0)
                else:
                    a = ma.masked_where(a == self.__dict__['_FillValue'],a,copy=0)
                a.set_fill_value(self.__dict__['_FillValue'])
        elif self.__dict__.has_key('missing_value'):
            if a.__contains__(self.__dict__['missing_value']):
                if np.isscalar(a):
                    a = ma.masked_where(a == self.__dict__['missing_value'][0],a,copy=0)
                else:
                    a = ma.masked_where(a == self.__dict__['missing_value'],a,copy=0)
                a.set_fill_value(self.__dict__['missing_value'])
    else: 
        # Handles MaskedIfFillAtt and MaskedAlways
        if self.__dict__.has_key('_FillValue'):
            if np.isscalar(a):
                a = ma.masked_where(a == self.__dict__['_FillValue'][0],a,copy=0)
            else:
                 a = ma.masked_where(a == self.__dict__['_FillValue'],a,copy=0)
            a.set_fill_value(self.__dict__['_FillValue'])
        elif self.__dict__.has_key('missing_value'):
            if np.isscalar(a):
                a = ma.masked_where(a == self.__dict__['missing_value'][0],a,copy=0)
            else:
                a = ma.masked_where(a == self.__dict__['missing_value'],a,copy=0)
            a.set_fill_value(self.__dict__['missing_value'])
        elif self.file.ma_mode == 'maskedalways':
            # supply a mask of all False, but just allow the fill_value to default
            mask = np.zeros(a.shape,dtype='?')
            a = ma.array(a,mask=mask,copy=False) 

    return a
   
def __getitem__(self, xsel):
    ''' 
Return data as specified by selection object xsel from an NioVariable. 
Depending on the setting of option MaskedArrayMode and the presence of 
a fill value attribute and/or fill values in the data, the return value will
be a MaskedArray or a normal NumPy array.
    '''

    ret = get_variable(self.file, self.varname, xsel)

    ret = _fill_value_to_masked(self,ret)

    return ret

def get_value(self):
    '''
Retrieve the value of a variable in the file.

v = f.variables['varname']
val = v.get_value()

'val' is returned as a NumPy array or a MaskedArray depending on the value
of option MaskedArrayMode. This method is the only way to retrieve the value
of a scalar variable in a file. For array variables basic or extended selection
syntax is more flexible.
    '''
    
    ret  = self._obj.get_value()
    ret = _fill_value_to_masked(self,ret)

    return ret

def _masked_to_fill_value(self,value):
    fill_value = None
    add_fill_value_att = False

    if ma.isMaskedArray(value):
        #
        # If the file variable already has a _FillValue or missing_value attribute
        # use it for the fill_value when converting the masked array.
        # _FillValue is the preferred fill value attribute but if it is not set
        # then look for missing_value.
        # Note that it is important to generate the fill_value with the correct type.
        # If there is an existing fill value in the file, make it conform to that type (??)
        # Otherwise use the type of the numpy array value
        #
        if self.__dict__.has_key('_FillValue'):
            fval = self.__dict__['_FillValue']
            fill_value = np.array(fval,dtype=fval.dtype)
        elif self.__dict__.has_key('missing_value'):
            fval = self.__dict__['missing_value']
            fill_value = np.array(fval,dtype=fval.dtype)
        elif _is_new_ma:
            fill_value = np.array(value.fill_value,dtype=value.dtype)
            add_fill_value_att = True
        else:
            fill_value = np.array(value.fill_value(),dtype=value.dtype)
            add_fill_value_att = True
        value = value.filled(fill_value)

    if add_fill_value_att:
        setattr(self._obj,'_FillValue',fill_value)

    return value

def __setitem__(self, xsel, value):
    ''' 
Assign elements of value to file variable with subscripts specified by the 
selection object xsel. If the value is a masked array fill it using the file 
variable fill value if it exists; otherwise use the masked array fill value.
    '''

    xsel = inp2xsel(self.file, self.varname, xsel)

    value = _masked_to_fill_value(self,value)

    if not isinstance(xsel, xSelect) or xsel.isbasic:
        self._obj[xsel] = value
    else:
        bb = xsel.bndbox()
        rsel = xsel - bb
        ret = self._obj[bb]
        ret[rsel] = value

    return

def assign_value(self,value):
    '''
Assign a value to a variable in the file.

v = f.variables['varname']
v.assign_value(value)
value - a NumPy array, a MaskedArray, or a Python sequence of values 
that are coercible to the type of variable 'v'.

This method is the only way to assign to a scalar variable in the file. 
For array variables basic or extended selection syntax is more flexible.
    '''

    value = _masked_to_fill_value(self,value)
    self._obj.assign_value(value)

    return

import sys
def close(self,history=""):
    '''
Close a file, ensuring all modifications are up-to-date if open for writing.

f.close([history])
history -- optional string appended to the global 'history' attribute
before closing a writable file. The attribute is created if it does not
already exist.
Read or write access attempts on the file object after closing
raise an exception.
    '''
    #print "in close"
    if not self.parent is None:
        return
    if self.open == 1:
        self._obj.close(history)
        self.open = 0
    return


def create_variable(self,name,type,dimensions):
    '''
Create a new variable with given name, type, and dimensions in a writable file.

f.create_variable(name,type,dimensions)
name -- a string specifying the variable name.
type -- a type identifier. The following are currently supported:
    'd' -- 64 bit real
    'f' -- 32 bit real
    'i' -- 32 bit integer
    'I' -- 32 bit unsigned integer
    'q' -- 64 bit integer
    'Q' -- 64 bit unsigned integer
    'l' -- 32 or 64 bit integer (platform dependent)
    'L' -- 32 or 64 bit unsigned integer (platform dependent)
    'q' -- 64 bit integer
    'Q' -- 64 bit unsigned integer
    'h' -- 16 bit integer
    'H' -- 16 bit unsigned integer
    'b' -- 8 bit integer
    'B' -- 8 bit unsigned integer
    'S1','c' -- character
dimensions -- a tuple of dimension names (strings), previously defined
    '''

    #print 'in create variable'
    v = self._obj.create_variable(name,type,dimensions)
    if not v is None:
        vp  = _proxy(v,'str','len',
                     __setitem__=__setitem__,__getitem__=__getitem__,get_value=get_value,assign_value=assign_value)
        vp.file = weakref.proxy(self)
        vp.varname = name
        vp.cf_dimensions = vp.dimensions
        self.variables[name] = vp
        if not self.variables.topdict is None:
            self.variables.topdict[name] = vp
        return vp
    else:
        return None

def create_group(self,name):
    '''
Create a new group with given name in a writable file.

f.create_group(name)
name -- a string specifying the group name.
    '''

    #print 'in create group'
    #import pdb; pdb.set_trace()
    g = self._obj.create_group(name)
    if not g is None:
        gp  = _proxy(g,'str')
        #gp.file = weakref.proxy(g)
        gp.file = g
        if self.parent is None:
            gp.parent = weakref.proxy(self.groups['/'])
        else:
            gp.parent = weakref.proxy(self)
        gp.ma_mode = self.ma_mode
        gp.groups = nd.nioDict()
        gp.groups.path = g.path
        gp.variables = nd.nioDict()
        gp.variables.path = g.path
        if self.groups.topdict is None:
            self.groups[name] = gp
            self.groups['/'].groups[name] = gp
            gp.groups.topdict = self.groups
            gp.variables.topdict = self.variables
        elif self.name is '/': 
            self.groups.topdict[name] = gp
            self.groups[name] = gp
            gp.groups.topdict = self.groups.topdict
            gp.variables.topdict = self.variables.topdict
        else:
            path = self.path + '/' + name
            self.groups.topdict[path] = gp
            self.groups[name] = gp
            gp.groups.topdict = self.groups.topdict
            gp.variables.topdict = self.variables.topdict
        return gp
    else:
        return None

def _get_masked_array_mode(options,option_defaults):
    ''' 
        get the MaskedArrayMode value considering the default setting and any option value set 
        when the file is opened 
    '''

    # ma_mode specifies when to return masked arrays
    # MaskedIfFillAtt: return a masked array iff file variable has a _FillValue or a missing_value (default)
    # MaskedNever: never return a masked array for any variable
    # MaskedAlways: return a masked array for all variables
    # MaskedIfFillAttAndValue: return a masked array iff file variable has a _FillValue or a missing_value and
    #                          the returned data array actually contains 1 or more fill values.
    # MaskedExplicit: only mask values specified explicitly using options ExplicitFillValues, MaskBelowValue,
    #                 and MaskAboveValue; ignore fill value attributes associated with the variable

    optvals = [ 'maskednever', 'maskediffillatt', 'maskedalways', 'maskediffillattandvalue', 'maskedexplicit' ]

    if options is not None:
        for key in options.__dict__.keys():
            lkey = key.lower()
            if not lkey == 'maskedarraymode':
                continue
            val = options.__dict__[key]
            lval = val.lower()
            if optvals.count(lval) == 0:
                raise ValueError, 'Invalid value for MaskArrayMode option'
            return  lval

    if option_defaults.has_key('MaskedArrayMode'):
        return option_defaults['MaskedArrayMode'].lower()
    else:
        return 'maskediffillatt'

def _get_axis_att(options,option_defaults):
    ''' Get a value for the UseAxisAttribute option '''
    if options == None:
        if option_defaults.has_key('UseAxisAttribute'):
            return option_defaults['UseAxisAttribute']
        else:
            return False
    for key in options.__dict__.keys():
        lkey = key.lower()
        if not lkey == 'useaxisattribute':
            continue
        val = options.__dict__[key]
        if val:
            return True
        return False
    return False

def _get_option_value(options,option_defaults,name):
    ''' Get an option value when the file is opened, considering 
        the default value and and any value set using  options keyword.
        The name must be one of the valid Python-level options and 
        have a default value of None.
    '''

    if options is not None:
        lname = name.lower()
        for key in options.__dict__.keys():
            lkey = key.lower()
            if not lkey == lname:
                continue
            val = options.__dict__[key]
            if val:
                return val
            return None

    if option_defaults.has_key(name):
        return option_defaults[name]
    else:
        return None

def set_option(self,option,value):
    '''
    Set certain options for an open NioFile instance.
    The option name is specified as a string
    Options that can be set include:
    MaskedArrayMode -- Specify MaskedArray bahavior (string):
        'MaskedIfFillAtt' -- Return a masked array iff file variable has a 
            _FillValue or a missing_value attribute (default).
        'MaskedNever' -- Never return a masked array for any variable.
        'MaskedAlways' -- Return a masked array for all variables.
        'MaskedIfFillAttAndValue' -- Return a masked array iff file variable has a 
            _FillValue or a missing_value attribute and the returned data array 
            actually contains 1 or more fill values.
        'MaskedExplicit' -- Only mask values specified explicitly using options 
            'ExplicitFillValues, MaskBelowValue, and/or MaskAboveValue; 
            ignore fill value attributes associated with the variable.
    ExplicitFillValues -- A scalar value or a sequence of values to be masked in the 
        return array. The first value becomes the fill_value attribute of the MaskedArray.
        Setting this option causes MaskedArrayMode to be set to 'MaskedExplicit'.
    MaskBelowValue -- A scalar value all values less than which are masked. However, if
        MaskAboveValue is less than MaskBelowValue, a range of of values become masked.
        Setting this option causes MaskedArrayMode to be set to 'MaskedExplicit'.
    MaskAboveValue -- A scalar value all values greater than which are masked. However, if
        MaskBelowValue is greater than MaskAboveValue, a range of of values become masked.
        Setting this option causes MaskedArrayMode to be set to 'MaskedExplicit'.
    '''
    valid_opts = {'maskedarraymode':'ma_mode', 'explicitfillvalues':'explicit_fill_values',
                  'maskbelowvalue' : 'mask_below_value', 'maskabovevalue' : 'mask_above_value' }
    if hasattr(option,'lower'):
        loption = option.lower()
    else:
        loption = option
    if not valid_opts.has_key(loption):
        raise KeyError, "Option %s invalid or cannot be set on open NioFile" % (option,)
        
    if hasattr(value,'lower'):
        lvalue = value.lower()
    else:
        lvalue = value
    if loption == 'explicitfillvalues' or loption == 'maskbelowvalue' or loption == 'maskabovevalue':
        if lvalue is not None:
            setattr(self,'ma_mode','maskedexplicit')

    return setattr(self,valid_opts[loption],lvalue)

#for debugging only
#import sys


def __del__(self):
    #print "in __del__ method"
    if hasattr(self,"parent") and self.parent is not None:
        return None
    if hasattr(self,'open') and self.open == 1:
        self.close()
        self.open = 0
    del(self)
    return None

'''
    if self.groups is not None:
        for gname in self.groups:
            grp = self.groups[gname]
            grp.variables.name = None
            grp.variables.parent = None
#            grp.variables._obj = None
            grp.variables.file = None
            grp.variables = None
            grp.groups = None
            grp.file = None
            grp.parent = None
    self.variables.name = None
    self.variables.parent = None
 #   self.variables._obj = None
    self.variables.file = None
    self.variables = None
    self.groups.file = None
    self.groups.parente = None
    self.file = None
    self.parent = None
    
    return None
'''

    
def open_file(filename, mode = 'r', options=None, history='', format=''):
    '''
Open a file containing data in a supported format for reading and/or writing.

f = Nio.open_file(filepath, mode='r',options=None, history='', format='')
filepath -- path of file with data in a supported format. Either the path must
end with an extension indicating the expected format of the file (it need not
be part of the actual file name), or it must be specified using the optional 
'format' argument. Valid extensions include:
    .nc, .cdf, .netcdf, .nc3, .nc4,  -- NetCDF
    .gr, .grb, .grib, .gr1, .grb1, .grib1, .gr2, .grb2, .grib2, -- GRIB
    .hd, .hdf -- HDF
    .h5, .hdf5 -- HDF5
    .he2, .he4, .hdfeos -- HDFEOS
    .he5, .hdfeos5 -- HDFEOS5
    .shp, .mif, .gmt, .rt1 -- shapefiles, other formats supported by GDAL OGR
    .ccm -- CCM history files
Extensions are handled case-insensitvely, i.e.: .grib, .GRIB, and .Grib all
indicate a GRIB file.
mode -- access mode (optional):
     'r' -- open an existing file for reading
     'w','r+','rw','a' -- open an existing file for modification
     'c' -- create a new file open for writing
options -- instance of NioOptions class used to specify generic or 
format-specific options.
history -- a string specifying text to be appended to the file's global
attribute. The attribute is created if it does not exist. Only valid
if the file is opened for writing.
format -- a string specifying the expected format. Valid strings are the
same as the extensions listed above without the initial period (.). 

Returns an NioFile object.
    '''

    ma_mode  = _get_masked_array_mode(options,_Nio.option_defaults)
    use_axis_att = _get_axis_att(options,_Nio.option_defaults)
    explicit_fill_values = _get_option_value(options,_Nio.option_defaults,'ExplicitFillValues')
    mask_below_value = _get_option_value(options,_Nio.option_defaults,'MaskBelowValue')
    mask_above_value = _get_option_value(options,_Nio.option_defaults,'MaskAboveValue')
    
    file = _Nio.open_file(filename,mode,options,history,format)

    file_proxy = _proxy(file, 'str', __del__=__del__,create_variable=create_variable,create_group=create_group,close=close)
    setattr(file_proxy.__class__,'set_option',set_option)
    file_proxy.file = file
    file_proxy.parent = None
    file_proxy.open = 1
    if not (explicit_fill_values is None and mask_below_value is None and mask_above_value is None):
        file_proxy.ma_mode = 'maskedexplicit'
    else:
        file_proxy.ma_mode = ma_mode
    file_proxy.explicit_fill_values = explicit_fill_values
    file_proxy.mask_below_value = mask_below_value
    file_proxy.mask_above_value = mask_above_value

    if use_axis_att:
        cf_dims = _get_cf_dims(file_proxy)
        newdims = {}
        cf2dims = {}
        dimensions = file_proxy.dimensions
        for dim in dimensions:
            try:
                newdim = cf_dims[dim]
            except KeyError:
                newdim = dim
            newdims[newdim] = dimensions[dim]
            cf2dims[newdim] = dim
    else:
        cf2dims = None
        newdims = file_proxy.dimensions
    file_proxy.cf_dimensions = newdims
    file_proxy.cf2dims = cf2dims

    variable_proxies = nd.nioDict()
    variable_proxies.path = '/'
    variable_proxies.topdict = None
    for var in file.variables.keys():
        # print var, ": ", sys.getrefcount(file.variables[var])
        vp  = _proxy(file.variables[var],'str','len',
                     __setitem__=__setitem__,__getitem__=__getitem__,get_value=get_value,assign_value=assign_value)
        vp.file = weakref.proxy(file_proxy)
#        vp.file = file_proxy
        vp.varname = var
        vp.parent = None
        variable_proxies[var] = vp
        if use_axis_att:
            newdims = []
            dimensions = vp.dimensions
            for dim in dimensions:
                try:
                    newdim = cf_dims[dim]
                except KeyError:
                    newdim = dim
                newdims.append(newdim)
            vp.cf_dimensions = tuple(newdims)
        else:
            vp.cf_dimensions = vp.dimensions
        #print var, ": ", sys.getrefcount(file.variables[var])
    file_proxy.variables = variable_proxies

    group_proxies = nd.nioDict()
    group_proxies.path = '/'
    group_proxies.topdict = None
    for group in file.groups.keys():
        # print group, ": ", sys.getrefcount(file.groups[group])
        gp = _proxy(file.groups[group], 'str')
        gp.file = weakref.proxy(file.groups[group])
        gp.ma_mode = file_proxy.ma_mode
        #gp.file = file_proxy
        group_proxies[group] = gp

        v_proxy_refs = nd.nioDict()
        v_proxy_refs.path = file.groups[group].path
        v_proxy_refs.topdict = weakref.proxy(file_proxy.variables)
        #v_proxy_refs.topdict = file_proxy.variables
        for var in file.groups[group].variables.keys():
            #print file.groups[group].variables[var].path
            vp = file_proxy.variables[file.groups[group].variables[var].path]
            #print vp is file_proxy.variables[file.groups[group].variables[var].path]
            vp.parent = weakref.proxy(gp)
            #vp.parent = gp
            v_proxy_refs[var] = weakref.proxy(vp)
            #v_proxy_refs[var] = vp
        gp.variables = v_proxy_refs
        #print group, ": ", sys.getrefcount(file.groups[group])

    file_proxy.groups = group_proxies
    # all the groups need proxies before we can do this

    for group in file.groups.keys():
        g = file.groups[group]
        gp = file_proxy.groups[group]
        sl = group.rfind('/')
        if sl == -1:
            gp.parent = weakref.proxy(file_proxy.groups['/'])
            #gp.parent = file_proxy.groups['/']
        elif sl == 0:
            gp.parent =  weakref.proxy(file_proxy)
        else:
            gp.parent = weakref.proxy(file_proxy.groups[group[0:sl]])
            #gp.parent = file_proxy.groups[group[0:sl]]
        g_proxy_refs = nd.nioDict()
        g_proxy_refs.path = g.path
        g_proxy_refs.topdict = weakref.proxy(file_proxy.groups)
        #g_proxy_refs.topdict = file_proxy.groups
        for grp in g.groups.keys():
            grp_obj = g.groups[grp]
            #print grp_obj.path
            gproxy = file_proxy.groups[grp_obj.path]
            g_proxy_refs[grp] = gproxy
        gp.groups = g_proxy_refs



#    for var in file.variables.keys():
#        file_proxy.variables[var].coordinates = _create_coordinates_attriibute(file_proxy,file_proxy.variables[var])

    

    #print file_proxy, file_proxy.variables
    return file_proxy

def _create_coordinates_attriibute(file,var):
    coords = nd.nioDict()
    #print "creating coordinates attribute for ", var.name
    if len(var.dimensions) == 1 and var.dimensions[0] == var.name:
        coords[var.name] = var
        return coords

    if var.parent is None:
        parent = file
    else:
        parent = var.parent
    
    cnames = None
    cdict = {}
    if var.attributes.has_key('coordinates'):
        s = var.attributes['coordinates'].replace(',',' ')
        cnames = s.split()
        #print var.attributes['coordinates']

    dlist = []
    if not cnames is None:
        for c in cnames:
            if c in parent.variables.keys():
                cdict[c] = parent.variables[c]
                if dlist.count(cdict[c].dimensions) == 0:
                    dlist.append(cdict[c].dimensions)

    #if len(cdict) > 0: print cdict, dlist
    dims = []
    if len(dlist) == 1:  # means there is only one set of coordinates involved 
        cd = dlist[0] # cd is the dimension tuple
        dims = list(var.dimensions)
        for d in cd:
            if d in dims:
                dims.remove(d)
        coords[cd] = tuple(cdict.values())
                
    for d in dims:
        while not parent is None:
            if d in parent.dimensions.keys():
                for v in parent.variables.values():
                    if d == v.name and v.rank == 1:
                        coords[d] = v
                        break
            parent = parent.parent

    #print coords
    return coords


def _get_cf_dims(file):
    ret = {}
    for dim in file.dimensions:
        if dim in file._obj.variables:
            try:
                axis = file._obj.variables[dim].axis
                ret[dim] = axis.lower()
            except AttributeError:
                pass
    return ret
            
def options():
    '''
Return an NioOptions object for specifying format-specific options.
opt = Nio.options()
Assign 'opt' as the third (optional) argument to Nio.open_file.
print opt.__doc__ to see valid options.
    '''
    opt = _Nio.options()
    return opt

'''
Since the NioVariable and NioFile types do not actually get created until a file is opened, 
do a minimal file open here to create the types and make them public
'''

filepath = os.getenv("NCARG_NCARG")
fname = 'netcdf/pop.nc'
if filepath:
    filepath = os.path.join(filepath,"data",fname)
  
if os.access(filepath,os.R_OK):
    file = _Nio.open_file(filepath)
    file_proxy = _proxy(file, 'str', __del__=__del__,create_variable=create_variable,create_group=create_group,close=close)
    file_proxy.parent = None
    var_proxy = _proxy(file.variables['t'],'str','len',
                       __setitem__=__setitem__,__getitem__=__getitem__,get_value=get_value,assign_value=assign_value)

    for cls in _known_proxy_classes.values():
        if isinstance(file_proxy,cls):
            NioFile = cls
        elif isinstance(var_proxy,cls):
            NioVariable = cls

    __all__.append(NioFile)
    __all__.append(NioVariable)
    file.close()
    del(file_proxy)
    del(var_proxy)
    del(file)
    del os

 



#
# This script builds PyNIO from source. Some environment variables
# may be required. See the following comments.
#

# Test to make sure we actually have NumPy.
try:
  import numpy
except ImportError:
  print "Error: Cannot import NumPy. Can't continue."
  sys.exit()

#
# At a minimum, you must have NCL, NetCDF (3.6.0 or later), and 
# HDF-4 (4.1 or later) installed.
#
# You must set the environment variables:
#
#    NCARG_ROOT
#    NETCDF_PREFIX
#    HDF_PREFIX
#
# to the parent locations of the NCL, NetCDF-3, and HDF-4 installations.
#
# Note: at a minimum, you must have NCARG_ROOT set. If all the other
# software is installed in this same root directory, then you don't
# need to set any of the previous or following XXXX_PREFIX variables.
#
# You can optionally build PyNIO with NetCDF-4, HDF-EOS2, HDF-EOS5,
# GRIB2 and/or shapefile (using the GDAL library) support.  To do this, 
# the corresponding environment variables:
#
#    HAS_NETCDF4
#    HAS_HDFEOS
#    HAS_HDFEOS5
#    HAS_GRIB2
#    HAS_GDAL
#
# must be set to 1. In addition, the corresponding environment variables:
#
#    NETCDF4_PREFIX
#    HDFEOS_PREFIX
#    HDFEOS5_PREFIX
#    GRIB2_PREFIX 
#    GDAL_PREFIX
#
# must be set to the root location of that software, unless they are
# all the same as a previous setting, like NCARG_ROOT.
#
#
# If you are linking against NetCDF version 4, this script assumes
# that OPeNDAP support has been included since NetCDF 4.1.1 turns
# this on by default. If you built NetCDF-4 with OPeNDAP support
# turned off (--disable-dap), then set the environment variable
# HAS_OPENDAP to 0.
#
# If your HDF4 library was built with support for SZIP compression or
# if you want to include NETCDF4 and/or HDFEOS5 support and the HDF5 
# libraries on which they depend have SZIP support included, then you
# additionally need to set environment variables for SZIP in a similar
# fashion:
#    HAS_SZIP
# (set to 1)
#    SZIP_PREFIX
# (if it resides in a location of its own)
#
# Finally, you may need to include Fortran system libraries
# (like "-lgfortran" or "-lf95") to resolve undefined symbols.
#
# Use F2CLIBS and F2CLIBS_PREFIX for this. For example, if you
# need to include "-lgfortran", and this library resides in /sw/lib:
#
#  F2CLIBS gfortran
#  F2CLIBS_PREFIX /sw/lib
#

import os, sys, commands

#
# This proceure tries to figure out which extra libraries are
# needed with curl.
#
def set_curl_libs():
  curl_libs = commands.getstatusoutput('curl-config --libs')
  if curl_libs[0] == 0:
#
# Split into individual lines so we can loop through each one.
#
    clibs = curl_libs[1].split()
#
# Check if this is a -L or -l string and do the appropriate thing.
#
    for clibstr in clibs:
      if clibstr[0:2] == "-L":
        LIB_DIRS.append(clibstr.split("-L")[1])
      elif clibstr[0:2] == "-l":
        LIBRARIES.append(clibstr.split("-l")[1])
  else:
#
# If curl-config doesn't produce valid output, then try -lcurl.
#
    LIBRARIES.append('curl')

  return

# End of set_curl_libs

try:
  ncarg_root = os.environ["NCARG_ROOT"]
  LIB_DIRS   = [os.path.join(ncarg_root,'lib') ]
  INC_DIRS   = [os.path.join(ncarg_root,'include/ncarg/nio')]
  #LIB_DIRS   = ['libsrc']
  #INC_DIRS   = ['libsrc']

except:
  print "NCARG_ROOT is not set; can't continue!"
  sys.exit()

# These are the required NCL, HDF4, and NetCDF libraries.
LIBRARIES = ['nio', 'mfhdf', 'df', 'jpeg', 'png', 'z', 'netcdf']
# Check for XXXX_PREFIX environment variables.
try:
  LIB_DIRS.append(os.path.join(os.environ["NETCDF_PREFIX"],"lib"))
  INC_DIRS.append(os.path.join(os.environ["NETCDF_PREFIX"],"include"))
except:
  pass

try:
  LIB_DIRS.append(os.path.join(os.environ["HDF_PREFIX"],"lib"))
  INC_DIRS.append(os.path.join(os.environ["HDF_PREFIX"],"include"))
except:
  pass

try:
  HAS_HDFEOS5 = int(os.environ["HAS_HDFEOS5"])
  if HAS_HDFEOS5 > 0:
    LIBRARIES.append('he5_hdfeos')
    LIBRARIES.append('Gctp')
    try:
      LIB_DIRS.append(os.path.join(os.environ["HDFEOS5_PREFIX"],"lib"))
      INC_DIRS.append(os.path.join(os.environ["HDFEOS5_PREFIX"],"include"))
    except:
      pass
except:
  HAS_HDFEOS5 = 0

try:
  HAS_NETCDF4 = int(os.environ["HAS_NETCDF4"])
  if HAS_NETCDF4 > 0:
    LIBRARIES.append('hdf5_hl')
    LIBRARIES.append('hdf5')
    try:
      HAS_OPENDAP = int(os.environ["HAS_OPENDAP"])
    except:    
      HAS_OPENDAP = 1
    if HAS_OPENDAP > 0:
      set_curl_libs()
    try:
      LIB_DIRS.append(os.path.join(os.environ["NETCDF4_PREFIX"],"lib"))
      INC_DIRS.append(os.path.join(os.environ["NETCDF4_PREFIX"],"include"))
    except:
      pass
except:
  HAS_NETCDF4 = 0

try:
  HAS_HDFEOS = int(os.environ["HAS_HDFEOS"])
  if HAS_HDFEOS > 0:
    LIBRARIES.append('hdfeos')
    LIBRARIES.append('Gctp')
    try:
      LIB_DIRS.append(os.path.join(os.environ["HDFEOS_PREFIX"],"lib"))
      INC_DIRS.append(os.path.join(os.environ["HDFEOS_PREFIX"],"include"))
    except:
      pass
except:
  HAS_HDFEOS = 0

try:
  HAS_GRIB2 = int(os.environ["HAS_GRIB2"])
  if HAS_GRIB2 > 0:
    LIBRARIES.append('grib2c')
    LIBRARIES.append('jasper')   # png is needed again, b/c it 
    LIBRARIES.append('png')      # must come after jasper
    try:
      LIB_DIRS.append(os.path.join(os.environ["GRIB2_PREFIX"],"lib"))
      INC_DIRS.append(os.path.join(os.environ["GRIB2_PREFIX"],"include"))
    except:
      pass
except:
  HAS_GRIB2 = 0

try:
  HAS_GDAL = int(os.environ["HAS_GDAL"])
  if HAS_GRIB2 > 0:
    LIBRARIES.append('gdal')
    LIBRARIES.append('proj') 
    try:
      LIB_DIRS.append(os.path.join(os.environ["GDAL_PREFIX"],"lib"))
      INC_DIRS.append(os.path.join(os.environ["GDAL_PREFIX"],"include"))
    except:
      pass
except:
  HAS_GDAL = 0

try:
  HAS_SZIP = int(os.environ["HAS_SZIP"])
  if HAS_SZIP > 0:
    LIBRARIES.append('sz')
    try:
      LIB_DIRS.append(os.path.join(os.environ["SZIP_PREFIX"],"lib"))
      INC_DIRS.append(os.path.join(os.environ["SZIP_PREFIX"],"include"))
    except:
      pass
except:
  HAS_SZIP = 0

# Depending on what Fortran compiler was used to build, we may need
# additional library paths or libraries.
try:
  f2clibs = os.environ["F2CLIBS"].split()
  for lib in f2clibs:
    LIBRARIES.append(lib)
except:
  pass

try:
  LIB_DIRS.append(os.environ["F2CLIBS_PREFIX"])
except:
  pass

try:
  EXTRA_OBJECTS = [os.environ["EXTRA_OBJECTS"]]
except:
  EXTRA_OBJECTS = ""

#
# Done with environment variables.
#

#
# Function for getting list of needed GRIB2 code tables and
# copying them over for the PyNIO installation.
#
def get_grib2_codetables():
  plat_dir = os.path.join("build","lib."+get_platform()+"-"+sys.version[:3], \
                          "PyNIO")

  ncl_lib       = os.path.join(ncarg_root,'lib')
  ncl_ncarg_dir = os.path.join(ncl_lib,'ncarg')
  ncarg_dirs    = ["grib2_codetables"]

  cwd = os.getcwd()          # Retain current directory.
  if not os.path.exists('ncarg'):
    os.mkdir('ncarg')          # make a directory to copy files to
  os.chdir(ncl_ncarg_dir)    # cd to $NCARG_ROOT/lib/ncarg

# Walk through each directory and copy some data files.
  for ncarg_dir in ncarg_dirs:
    for root, dirs, files in os.walk(ncarg_dir):
      dir_to_copy_to = os.path.join(cwd,'ncarg',root)
      if not os.path.exists(dir_to_copy_to):
        os.mkdir(dir_to_copy_to)
      for name in files:
        file_to_copy = os.path.join(ncl_ncarg_dir,root,name)
        cmd = "cp " + file_to_copy + " " + dir_to_copy_to
        os.system(cmd)
        data_files.append(os.path.join('ncarg',root,name))

  os.chdir(cwd)    # cd back to original directory

  return


# Main code.

import platform
from distutils.core import setup, Extension
from distutils.util import get_platform
from distutils.sysconfig import get_python_lib

#
# Initialize some variables.
#
pynio_vfile = "pynio_version.py"         # PyNIO version file.
pkgs_pth    = get_python_lib()

if sys.platform == "linux2" and os.uname()[-1] == "x86_64" and \
    platform.python_compiler()[:5] == "GCC 4":
    LIBRARIES.append('gfortran')

elif sys.platform == "irix6-64":
    LIBRARIES.append('ftn')
    LIBRARIES.append('fortran')
    LIBRARIES.append('sz')

elif sys.platform == "sunos5":
    LIBRARIES.append('fsu')
    LIBRARIES.append('sunmath')

elif sys.platform == "aix5":
    os.putenv('OBJECT_MODE',"64")
    LIBRARIES.append('xlf90')
    
#----------------------------------------------------------------------
#
# Set some variables.
#
#----------------------------------------------------------------------
from numpy import __version__ as array_module_version

# I read somewhere that distutils doesn't update this file properly
# when the contents of directories change.

if os.path.exists('MANIFEST'): os.remove('MANIFEST')

pynio_pkg_name = 'PyNIO'
pynio_pth_file = ['Nio.pth']
DMACROS        =  [ ('NeedFuncProto','1'), ('NIO_LIB_ONLY', '1') ]

INC_DIRS.insert(0,numpy.get_include())

#----------------------------------------------------------------------
#
# Create version file that contains version and array module info.
#
#----------------------------------------------------------------------
if os.path.exists(pynio_vfile):
  os.system("/bin/rm -rf " + pynio_vfile)

pynio_version = open('version','r').readlines()[0].strip('\n')

vfile = open(pynio_vfile,'w')
vfile.write("version = '%s'\n" % pynio_version)
vfile.write("array_module = 'numpy'\n")
vfile.write("array_module_version = '%s'\n" % array_module_version)
vfile.write("python_version = '%s'\n" % sys.version[:5])
vfile.close()

#----------------------------------------------------------------------
#
# Here are the instructions for compiling the "nio.so" file.
#
#----------------------------------------------------------------------
print '====> Installing Nio to the "'+pynio_pkg_name+'" site packages directory.'

module1 = [Extension('nio',
                    define_macros = DMACROS,
                    include_dirs  = INC_DIRS,
                    libraries     = LIBRARIES,
                    library_dirs  = LIB_DIRS,
                    extra_objects = EXTRA_OBJECTS,
                    sources       = ['niomodule.c']
                    )]

data_files = []
if HAS_GRIB2 > 0:
  get_grib2_codetables()

setup (name         = 'PyNIO',
       version      = pynio_version,
       description  = 'Multi-format data I/O package',
       author       = 'David I. Brown',
       author_email = 'dbrown@ucar.edu',
       url          = 'http://www.pyngl.ucar.edu/Nio.shtml',
       long_description = '''
       Enables NetCDF-like access for NetCDF (rw), HDF (rw), HDFEOS (r), GRIB (r), and CCM (r) data files
       ''',
       package_dir  = { pynio_pkg_name : '.' },
       packages     = [ pynio_pkg_name ],
       ext_modules  = module1,
       ext_package  = pynio_pkg_name,
       package_data = { pynio_pkg_name : data_files },
       data_files   = [(pkgs_pth, pynio_pth_file)])

#
# Cleanup: remove the pynio_version.py file.
#
if os.path.exists(pynio_vfile):
  os.system("/bin/rm -rf " + pynio_vfile)




#@+leo-ver=4-thin
#@+node:schmidli.20080321230001.1:@thin xarray.py
#@@language python
#@<< xarray documentation >>
#@+node:schmidli.20080321230412:<< xarray documentation >>
"""xarray module 

Defines the class _xArray and some helper functions.

_xArray supports multilinear interpolation and masked indexing arrays
in addition to numpy's basic slicing and advanced selection mechanisms.

"""
#@-node:schmidli.20080321230412:<< xarray documentation >>
#@nl
#@<< xarray declarations >>
#@+node:schmidli.20080321230001.2:<< xarray declarations >>
import numpy as N
from numpy import asarray

__version__ = '0.1.1'
__all__ = ['_xArray', '_intp', '_rindex', '_rindex2', '_rline', '__version__']

#@-node:schmidli.20080321230001.2:<< xarray declarations >>
#@nl
#@+others
#@+node:schmidli.20080321230001.3:class _xArray
class _xArray(N.ndarray):
    """ Extends numpy's ndarray class to support multilinear 
    interpolation and masked indexing arrays. In other words the
    elements of the selection tuble may contain floating point 
    numbers and masked arrays.

    """
    #@  @+others
    #@+node:schmidli.20080321230001.4:__new__
    def __new__(subtype, data, dtype=None, copy=True):
        """
            Returns an _xArray object.
        """
        if isinstance(data, N.ma.MaskedArray):
            mask = data.mask
            masked = True
            fill_value = data._fill_value
            data = data.filled(fill_value)
        else:
            mask = False
            masked = False
            fill_value = None
        arr = N.array(data, dtype=dtype, copy=copy)
        shape = arr.shape
        ret = N.ndarray.__new__(subtype, shape, arr.dtype, buffer=arr)
        ndim = arr.ndim
        ret.masked = masked
        ret.mask = mask
        ret.fill_value = fill_value
        return ret
    #@-node:schmidli.20080321230001.4:__new__
    #@+node:schmidli.20080321230001.5:__array_finalize__
    def __array_finalize__(self, obj):
        """ """
        self.masked = False
        self.mask = False
        self.fill_value = None
    #@-node:schmidli.20080321230001.5:__array_finalize__
    #@+node:schmidli.20080321230001.6:__copy__
    def __copy__(self):
        """ """
        ret = N.ndarray.__copy__(self)
        ret.masked = self.masked
        ret.mask = self.mask
        ret.fill_value = self.fill_value
        return ret
    #@-node:schmidli.20080321230001.6:__copy__
    #@+node:schmidli.20080321230001.7:__getitem__
    def __getitem__(self, index):
        """x.__getitem__(y) <==> x[y]
        """
        ret = _intp(asarray(self), index)
        return ret
    #@nonl
    #@-node:schmidli.20080321230001.7:__getitem__
    #@+node:schmidli.20080321230001.8:__str__
    def __str__(self):
        """ informal string representation """

        if self.masked:
            r = 'Masked: '
        else:
            r = ''
        r += str(self.__array__())
        return r
    #@-node:schmidli.20080321230001.8:__str__
    #@+node:schmidli.20080321230001.9:__repr__
    def __repr__(self):
        """ full string representation """

        r = '_xArray(' + self.__str__() + ')'
        return r
    #@-node:schmidli.20080321230001.9:__repr__
    #@+node:schmidli.20080321230001.10:getA
    def getA(self):
        """ base array """
        return asarray(self)

    A = property(getA)
    #@nonl
    #@-node:schmidli.20080321230001.10:getA
    #@-others
#@nonl
#@-node:schmidli.20080321230001.3:class _xArray
#@+node:schmidli.20080321230001.12:_intp
def _intp(ar, sltup):
    """
        Returns the result of applying an extended selection
        tuple to the array. 

        In addition to numpy's basic slicing and advanced selection,
        masked indexing arrays and multi-linear interpolation are
        supported.

        Linear interpolation along a specific axis is triggered
        when the corresponding member of the selection tuple contains
        an object (scalar, list, array) with non-integer numbers.

        Examples, using shorthand notation:
            v = ar[5]
            v = ar[2.2,:,(4,5,6)]
            v = ar[[2.5,6.2],0,:]
            v = ar[:,[[2,2.2],[3,mv]],[[2],[3]],[[4,5]]]
    """
    if not isinstance(sltup, tuple):
        sltup = (sltup,)
    if len(sltup) != ar.ndim:
        raise ValueError, 'length of selection tuple does not match array dimension'

    # determine first axis which requires interpolation
    axis = None
    for i in range(len(sltup)):
        if isinstance(sltup[i], (N.ndarray, N.ma.MaskedArray)):
            if sltup[i].dtype in (N.single, N.double):
                axis = i
                break
        elif isinstance(sltup[i], slice):
            pass
        elif N.array(sltup[i]).dtype in (N.single, N.double):
            axis = i
            break

    # if no interpolation required: apply selection and quit
    if axis is None:
        if len(sltup) == 0:
            return ar[sltup]
        elif ismasked_tuple(sltup):
            sltup = list(sltup)
            axis_mask = 0
            for n in range(len(sltup)):
                if isinstance(sltup[n], N.ma.MaskedArray):
                    mask = sltup[n].mask
                    sltup[n] = sltup[n].filled(0)
                    axis_mask = n
            sltup = tuple(sltup)
            r = ar[sltup]
            #print "r.shape, mask.shape:", r.shape, mask.shape, axis_mask
            if axis_mask > 0:
                new_mask = N.zeros(r.shape, dtype='bool')
                sl = (None,)*axis_mask + (slice(None),)
                new_mask[:] = mask[sl]
                mask = new_mask
            if isinstance(r, N.ma.MaskedArray): 
                return N.ma.array(r, mask=mask | r.mask)
            else:
                return N.ma.array(r, mask=mask)
        else:
            return ar[sltup]

    # calculate indices and weights
    rj = sltup[axis]
    j = N.floor(rj).astype(N.int)
    j = N.clip(j, 0, ar.shape[axis]-2)
    jp = j+1
    f0j = rj-j
    f1j = 1-f0j
    sl = list(sltup)
    slp = list(sltup)
    sl[axis] = j
    slp[axis] = jp
    sl = tuple(sl); slp = tuple(slp)

    res = _intp(ar, sl)*f1j + _intp(ar, slp)*f0j
    return res
#@-node:schmidli.20080321230001.12:_intp
#@+node:schmidli.20080321230916:_rindex
def _rindex(ar, val, axis=0, ep=0.5, clip=True, round=False, squeeze=True):
    """
        Assuming that ar is sorted in ascending order along axis, this
        function returns real indices of where val would fit (using linear
        interpolation).

        ar[j,I]     where I is a multi-dimensional index
        val[nv]     values
        ep          width of extrapolation zone
        clip        True: clip resulting indices to lie within a specified
                    range, i.e. -ep <= rj <= (n-1)+ep
                    False: set rj.mask = rj < -ep or rj > (n-1)+ep
        res[nv,I]

print _rindex(N.arange(10), 5.5)
        [ 5.5]
print _rindex(N.arange(10), -1)
        [-0.5]
print _rindex(N.arange(10), [2,4.5])
        [ 2.   4.5]
    """

    # transpose input array if necessary
    if axis != 0:
        tr = N.arange(ar.ndim)
        tr[0] = axis
        tr[axis] = 0
        ar = ar.transpose(tr).copy()

    # bring input arrays into required shape
    # ar[nj,ni] and val[nv]
    ar_shape = list(ar.shape)
    nj = ar_shape[0]
    ni = ar.size/nj
    ar.shape = (nj,ni)
    is_scalar = N.isscalar(val)
    val_ndim = N.array(val).ndim
    val = N.atleast_1d(val)
    nv = val.size

    # determine the r-index
    if ar.shape[0] == 1:
        rj = N.zeros((1,ar.shape[1]), dtype=N.int)
    else:
        if ar[0,0] < ar[1,0]: # ascending order
            j1 = N.apply_along_axis(N.searchsorted, 0, ar, val).reshape(nv,ni)
            j1 = N.clip(j1, 1, nj-1)
            j0 = j1-1
            i = N.arange(ni).reshape(1,ni)
            rj = j0 + 1.0*(val.reshape(nv,1) - ar[j0,i])/(ar[j1,i] - ar[j0,i])
            if clip:
                rj = N.clip(rj, -ep, (nj-1)+ep)
            else:
                mask = (rj<-ep)|(rj>nj-1+ep)
                if mask.any():
                    rj = N.ma.array(rj, mask=mask)
        else: #descending order
            j1 = N.apply_along_axis(N.searchsorted, 0, ar[::-1], val).reshape(nv,ni)
            j1 = nj - j1 - 1
            j1 = j1.clip(0,nj-2)
            j0 = j1+1
            i = N.arange(ni).reshape(1,ni)
            rj = j0 - 1.0*(val.reshape(nv,1) - ar[j0,i])/(ar[j1,i] - ar[j0,i])
            if clip:
                rj = N.clip(rj, -ep, (nj-1)+ep)
            else:
                mask = (rj<-ep)|(rj>nj-1+ep)
                if mask.any():
                    rj = N.ma.array(rj, mask=mask)



    # back to multdimensional
    ar.shape = ar_shape
    ar_shape[0] = nv
    rj.shape = ar_shape

    # and transpose if necessary
    if axis != 0:
        if isinstance(rj, N.ma.MaskedArray):
            mask = rj.mask.transpose(tr)
            data = rj.filled(0.).transpose(tr).copy()
            rj = N.ma.array(data, mask=mask)
        else:
            rj = rj.transpose(tr).copy()

    if round:
        rj = N.round(rj).astype(N.int)

    if is_scalar and squeeze:
        rj_shape = list(rj.shape)
        del(rj_shape[axis])
        rj.shape = tuple(rj_shape)
        #if len(ar_shape) == 1:
        #    rj = rj[0]
    return rj

#@-node:schmidli.20080321230916:_rindex
#@+node:schmidli.20080322212349:_rindex2
def _rindex2(ar, val, axis=0, ep=0.5, clip=True, round=False):
    """
        As function _rindex, but returns a complete selection tuple.
    """

    val_ndim = N.array(val).ndim
    rj = _rindex(ar, val, axis=axis, ep=ep, clip=clip, round=round, 
                squeeze=False)

    # determine selection tuples for the remaining dimensions
    rj_shape = list(rj.shape)
    rtup = []
    for i in xrange(rj.ndim):
        idx = N.arange(rj.shape[i])
        idx_shape = N.ones(rj.ndim)
        idx_shape[i] = rj.shape[i]
        idx.shape = idx_shape
        rtup.append(idx)
    rtup[axis] = rj

    if val_ndim == 0:
        for i in xrange(rj.ndim):
            r_shape = list(rtup[i].shape)
            if i != axis: del(r_shape[axis])
            rtup[i].shape = r_shape
    return tuple(rtup)
#@-node:schmidli.20080322212349:_rindex2
#@+node:schmidli.20080325135032:_rline
def _rline(edges, delta=None, ns=None):
    """
        Returns a discretized polyline
        -> 

        edges[npts,ndim]    coordinates of the edge points of the 
                            line segments
        delta           length of a discretized segment
        ns              number of discretized segments
    """

    edges = N.array(edges)
    if edges.ndim != 2:
        raise ValueError('edges must be 2-dimensional')
    b = edges[1:,:] - edges[0:-1,:]
    nlines = b.shape[0]                     # number of line segments
    d = N.zeros(nlines, dtype=N.double)
    for i in range(nlines):
        d[i] = N.sqrt((b[i]**2).sum())
    length = d.sum()
    if delta is None: delta = 1
    if ns is None: ns = int(length/delta)   # number of discret segments

    # number of discretized segments per line segment
    n = N.zeros(nlines, dtype=N.int_)
    for i in range(nlines):
        n[i] = N.round(d[i]/length*ns)

    rj = N.zeros((n.sum()+1,edges.shape[1]), dtype=N.double)
    idx = N.zeros(nlines+1, dtype=N.int_)
    idx[1:] = n.cumsum()
    for i in xrange(nlines):
        t = N.linspace(0,1,n[i]+1)
        t.shape = (t.shape[0],1)
        rj[idx[i]:idx[i+1]+1,:] = edges[i,:] + b[i,:]*t
    return rj
#@-node:schmidli.20080325135032:_rline
#@+node:schmidli.20080321230001.13:ismasked_tuple
def ismasked_tuple(sltup):
    """ """
    for n in range(len(sltup)):
        if isinstance(sltup[n], N.ma.MaskedArray):
            return True
    return False
#@-node:schmidli.20080321230001.13:ismasked_tuple
#@-others
#@-node:schmidli.20080321230001.1:@thin xarray.py
#@-leo




#@+leo-ver=4-thin
#@+node:schmidli.20080322120238:@thin pymfio/coordsel.py
#@@language python
#@<< coordsel declarations >>
#@+node:schmidli.20080322120238.1:<< coordsel declarations >>
""" Coordinate space selection for NioVariable

    data = f.variables[varname][selobj]

    where selobj is one of the following:
        input string (inp)
        coordinate space selection object (csel)
        index space selection object (isel)
        index space selection object (correct dimensionality; xsel)
        a basic numpy slice object
"""

import copy
import datetime
import time
import numpy as N
import Nio
from _xarray import _intp, _rindex, _xArray

__version__ = '0.1.0'
__all__ = ['get_variable', 'inp2csel', 'inp2isel', 'inp2xsel', 'idxsel2xsel', \
        'xSelect', 'crdSelect', 'idxSelect', '__version__']

fact = {'k': 10**3, 'M': 10**6, 'h': 3600, 'm': 60, 'H': 100}
fact_keys = fact.keys()
#@-node:schmidli.20080322120238.1:<< coordsel declarations >>
#@nl
#@+others
#@+node:schmidli.20080322120238.2:get_variable
def get_variable(file, varname, xsel):
    """ get variable from file using extended selection """

    dims = file.variables[varname].cf_dimensions
    order = []
    do_transpose = False
    if isinstance(xsel, str):
        xsel = crdSelect(xsel, dims)
        if hasattr(xsel,'names'):
            tdims = list(dims) 
            for name in xsel.names:
                order.append(tdims.index(name))
            for i in xrange(len(order)-1):
                if order[i] > order[i+1]:
                    do_transpose = True
            if not do_transpose:
                order = []
    if isinstance(xsel, crdSelect):
        xsel = crdsel2idxsel(file, xsel)
    if isinstance(xsel, idxSelect):
        xsel = idxsel2xsel(file, xsel, dims, order)

    if hasattr(xsel,'order'):
        order = xsel.order
    if not isinstance(xsel, xSelect) or xsel.isbasic:
        ret = file.file.variables[varname][xsel]
    else:
        if xsel.masked:
            ret = file.file.variables[varname][:]
            ret = _xArray(ret)[xsel]
        else:
            bb = xsel.bndbox()
            ret = file.file.variables[varname][bb]
            rsel = xsel - bb
            ret = _intp(ret, rsel)

    if do_transpose and len(order) > 1:
        ret = ret.transpose(order)

    return ret
#@-node:schmidli.20080322120238.2:get_variable
#@+node:schmidli.20080322120238.3:inp2csel
def inp2csel(file, varname, csel):
    """ Convert the input string to a coordinate space selection object. """

    dims = file.variables[varname].cf_dimensions
    if isinstance(csel, str):
        csel = crdSelect(csel, dims)

    return csel
#@-node:schmidli.20080322120238.3:inp2csel
#@+node:schmidli.20080322135237:inp2isel
def inp2isel(file, varname, isel):
    """ Convert the input isel to a index space selection object. """

    dims = file.variables[varname].cf_dimensions
    if isinstance(isel, str):
        isel = crdSelect(isel, dims)
    if isinstance(isel, crdSelect):
        isel = crdsel2idxsel(file, isel)

    return isel
#@-node:schmidli.20080322135237:inp2isel
#@+node:schmidli.20080322135142:inp2xsel
def inp2xsel(file, varname, xsel):
    """ Convert the input xsel to a index space selection object. """

    dims = file.variables[varname].cf_dimensions
    if isinstance(xsel, str):
        xsel = crdSelect(xsel, dims)
    if isinstance(xsel, crdSelect):
        xsel = crdsel2idxsel(file, xsel)
    if isinstance(xsel, idxSelect):
        xsel = idxsel2xsel(file, xsel, dims,[])

    return xsel
#@-node:schmidli.20080322135142:inp2xsel
#@+node:schmidli.20080322120238.4:class crdSelect
class crdSelect(dict):
    """ crdSelect(inp)

    Create a coordinate space selection object from inp

    Examples:
        crdSelect('time|d20070321-6:18:3h xc|10k:20k:2k')
    """
    #@  @+others
    #@+node:schmidli.20080322120238.5:__init__
    def __init__(self, inp, dimensions):
        """ crdSelect(inp, dimensions) """
        self.names = []
        if isinstance(inp, str):
            inp = inp.strip()
            inpv = inp.split()
            data = {}
            if '|' in inp:      # named axes input
                for item in inpv:
                    itemv = item.split('|')
                    if len(itemv) < 2:
                        raise ValueError, "Invalid input format"
                    key = itemv[0]
                    value = '|'.join(itemv[1:])
                    if dimensions is not None:
                        if key in dimensions:
                            data[key] = axisSelect(value)
                            self.names.append(key)
                        else:
                            pass
                            #raise ValueError, '"' + key + '" is not a valid axis name (UseAxisAttribute option setting?)'
                    else:
                        data[key] = axisSelect(value)
                        self.names.append(key)
            else:               # positional input
                if dimensions is None:
                    raise TypeError, "Invalid type for dimensions"
                if len(inpv) > len(dimensions):
                    inpv = inpv[:len(dimensions)]
                for i in range(len(inpv)):
                    key = dimensions[i]
                    data[key] = axisSelect(inpv[i])
                    self.names.append(key)
                if len(inpv) < len(dimensions):
                    for i in range(len(inpv), len(dimensions)):
                        key = dimensions[i]
                        data[key] = axisSelect('i:')
                        self.names.append(key)

            dict.__init__(self, data)
        else:
            raise TypeError, "Invalid input type"
    #@-node:schmidli.20080322120238.5:__init__
    #@+node:schmidli.20080322120238.6:__str__
    def __str__(self):
        """ string representation """
        _str = 'crdSelect(( '
        for key in self.keys():
            _str += str(key) +': ' + str(self[key]) + ', '
        _str = _str[:-2] + ' ))'

        return(_str)
    #@-node:schmidli.20080322120238.6:__str__
    #@-others
#@-node:schmidli.20080322120238.4:class crdSelect
#@+node:schmidli.20080322120238.7:class idxSelect
class idxSelect(dict):
    """ idxSelect(inp)

    Create a index space selection object from inp

    Examples:
        idxSelect(dimensions)
    """

    #@  @+others
    #@+node:schmidli.20080322120238.8:__init__
    def __init__(self, dimensions):
        """ idxSelect(dimensions) """

        if not isinstance(dimensions, list):
            raise TypeError, "Invalid argument type"
        data = {}
        for key in dimensions:
            data[key] = axisSelect('i:')

        self = dict.__init__(self, data)
    #@-node:schmidli.20080322120238.8:__init__
    #@+node:schmidli.20080322120238.9:__str__
    def __str__(self):
        """ string representation """

        _str = 'idxSelect(( '
        for key in self.keys():
            _str += str(key) +': ' + str(self[key]) + ', '
        _str = _str[:-2] + ' ))'

        return(_str)
    #@-node:schmidli.20080322120238.9:__str__
    #@-others
#@-node:schmidli.20080322120238.7:class idxSelect
#@+node:schmidli.20080322120238.10:class xSelect
class xSelect(tuple):
    """ xSelect(inp)

    Create an extended selection object to be used with _xArray.

    """

    #@  @+others
    #@+node:schmidli.20080322120238.11:bndbox
    def bndbox(self):
        """ Return the bounding box of the current selection object. """
        ret = []
        for idx in self:
            if isinstance(idx, slice):
                start = idx.start
                stop = idx.stop
            elif isinstance(idx, N.ndarray):
                start = idx.min()
                stop = idx.max()
            else:
                idx = N.atleast_1d(idx)
                start = idx.min()
                stop = idx.max()
            if isinstance(start, float):
                start = N.floor(start).astype(N.int)
            if isinstance(stop, float):
                stop = N.ceil(stop).astype(N.int)
            if not isinstance(idx, slice):
                stop += 1
            # should be true only for interpolation cases, therefore stop>= 2
            if start < 0: 
                start = 0
                stop = max(2, stop)

            ret.append(slice(start, stop))
        return tuple(ret)

        # does this do anything?
        bb = property(bndbox)   
    #@-node:schmidli.20080322120238.11:bndbox
    #@+node:schmidli.20080322120238.12:__sub__
    def __sub__(self, bb):
        """ Subtract the bounding box (bb) from the current selection object. """
        rsel = list(self)
        if len(rsel) != len(bb):
            raise ValueError, "Incompatible operands"
        for i in range(len(rsel)):
            if isinstance(self[i], slice):
                rsel[i] = slice(self[i].start-bb[i].start, \
                            self[i].stop-bb[i].start, self[i].step)
            else:
                rsel[i] = self[i] - bb[i].start
        return tuple(rsel)
    #@-node:schmidli.20080322120238.12:__sub__
    #@+node:schmidli.20080322120238.13:__str__
    def __str__(self):
        """ string representation """
        _str = 'xSelect('
        for item in self:
            _str += str(item) + ', '
        _str = _str[:-2] + '; isbasic: ' +str(self.isbasic)+ ')'

        return(_str)
    #@-node:schmidli.20080322120238.13:__str__
    #@-others
#@-node:schmidli.20080322120238.10:class xSelect
#@+node:schmidli.20080322120238.14:crdsel2idxsel
def crdsel2idxsel(file, csel):
    """ convert a coordinate space selection object to an index space
    selection object
    """

    isel = idxSelect(csel.keys())
    mdcrdname = {}
    for axis in isel.keys():
        if csel[axis].mdcrd is None:
            isel[axis] = csel[axis].toindex(file, axis)
        else:
            isel[axis] = axisIdxSelect(slice(0, file.cf_dimensions[axis]))
            mdcrdname[axis] = csel[axis].mdcrd

    if len(mdcrdname) > 1:
        raise NotImplementedError, "More than one multi-dimensional coordinate are not yet supported"

    for axis in mdcrdname.keys():
        isel[axis] = csel[axis].toindex(file, axis, mdcrd=mdcrdname[axis], 
                                        isel=isel)
    return isel
#@-node:schmidli.20080322120238.14:crdsel2idxsel
#@+node:schmidli.20080322120238.15:idxsel2xsel
def idxsel2xsel(file, isel, dimensions, order):
    """ convert a index space selection object to an xSelect object
    """

    if not isinstance(isel, idxSelect):
        raise TypeError, 'wrong argument type'

    xsel = {}
    xsel_size = {}
    xsel_dims = {}
    isarray = False
    interp = False
    masked = False
    multidim = False

    i = 0
    for axis in dimensions:
        inc_i = True
        try:
            idx = isel[axis]
            if idx.interp: interp = True
            if idx.isarray: 
                isarray = True
                if idx.dims is not None: multidim = True
                if isinstance(idx.v, N.ma.MaskedArray): masked = True
            xsel_dims[axis] = idx.dims
            idx = idx.v
            if isinstance(idx, slice):
                dimsize = file.cf_dimensions[axis]
                res = [idx.start, idx.stop, idx.step]
                if (idx.step is not None and idx.step < 0):
                    if idx.start is None: res[0] = dimsize - 1
                    if idx.stop is None: res[1] = None
                else:
                    if idx.start is None: res[0] = 0 
                    if idx.stop is None: res[1] = dimsize
                    if idx.step is None: res[2] = 1
                xsel[axis] = slice(res[0], res[1], res[2])
            elif N.isscalar(idx):
                xsel[axis] = idx
                if len(order) > 0:
                    order.remove(i)
                    for val in order:
                        if val > i:
                            order[order.index(val)] = val - 1
                            inc_i = False
            else:
                #xsel[axis] = idx.copy()
                xsel[axis] = copy.copy(idx)
                if len(idx.shape) == 0 or idx.shape == 1:
                    if len(order) > 0:
                        order.remove(i)
                        for val in order:
                            if val > i:
                                order[order.index(val)] = val - 1
                                inc_i = False
        except KeyError, e:
            dimsize = file.cf_dimensions[axis]
            xsel[axis] = (slice(0, dimsize, 1))
            xsel_dims[axis] = None
        if inc_i:
            i += 1

    if isarray:
        # convert slices to 1d-arrays and determine result size
        for axis in dimensions:
            idx = xsel[axis]
            if isinstance(idx, slice):
                xsel[axis] = N.arange(idx.start, idx.stop, idx.step)
            if xsel_dims[axis] is None:
                if is_scalar(xsel[axis]):
                    xsel_size[axis] = 0
                else:
                    xsel_size[axis] = len(xsel[axis])
            else:
                xsel_size[axis] = isel[axis].axlen

        # determine shape of xsel
        dim_ret = []
        for axis in dimensions:
            if xsel_size[axis] != 0:
                dim_ret.append(xsel_size[axis])
        ndim_ret = len(dim_ret)

        # all 1d arrays
        if not multidim:
            i = 0
            for axis in dimensions:
                if xsel_size[axis] != 0:
                    idx_shape = N.ones(ndim_ret,dtype="int32")
                    idx_shape[i] = dim_ret[i]
                    xsel[axis].shape = idx_shape
                    i += 1
        # at least one multidimensional coordinate
        else:
            i = 0
            for axis in dimensions:
                if xsel_dims[axis] is None:
                    if xsel_size[axis] != 0:
                        idx_shape = N.ones(ndim_ret,dtype="int32")
                        idx_shape[i] = dim_ret[i]
                        xsel[axis].shape = idx_shape
                        i += 1
                else:
                    idx_shape2 = {}
                    for axis2 in dimensions:
                        if xsel_size[axis2] != 0:
                            if axis2 in xsel_dims[axis]:
                                idx_shape2[axis2] = isel[axis].dimsize(axis2)
                            else:
                                idx_shape2[axis2] = 1
                    idx_shape = []
                    for axis2 in dimensions:
                        if axis2 in idx_shape2:
                            idx_shape.append(idx_shape2[axis2])
                    if isel[axis].type != 'scalar': 
                        i += 1

    # check if we only need basic slicing
    if not isarray and not interp:
        isbasic = True
    else:
        isbasic = False

    ret = []
    for axis in dimensions:
        ret.append(xsel[axis])
    ret = xSelect(ret)
    ret.isbasic = isbasic
    ret.interp = interp
    ret.masked = masked
    ret.order = order

    return ret            
#@-node:schmidli.20080322120238.15:idxsel2xsel
#@+node:schmidli.20080322120238.16:numpy2xsel
def numpy2xsel(isel):
    """ convert a numpy selection object to an xselection object

        extended numpy selection object:
            if multidim: the dimensionality of idx is NOT changed
            else: convert 1d-idx to ndim arrays
    """

    if N.isscalar(isel): isel = tuple((isel,))
    if isinstance(isel, slice): isel = tuple((isel,))
    xsel = []
    isarray = False
    interp = False
    multidim = False
    do_convert = True

    if not isinstance(isel, tuple):
        raise TypeError, "wrong argument type"

    for idx in isel:
        if isinstance(idx, slice):
            xsel.append(idx)
        elif N.isscalar(idx):
            xsel.append(idx)
            if idx.dtype in (N.float32, N.float64):
                interp = True
        else:
            isarray = True
            idx = N.atleast_1d(idx)
            xsel.append(idx)
            isarray = True
            if idx.dtype in (N.float32, N.float64):
                interp = True
            if idx.ndim > 1: 
                multidim = True      # conversion not supported for multidim

    # convert selection objects to compatible _intp arrays if necessary
    if isarray and not multidim:
        # convert slices to 1d-arrays
        for i in xrange(len(xsel)):
            if isinstance(xsel[i], slice):
                xsel[i] = N.arange(xsel[i].start, xsel[i].stop, xsel[i].step)
                xsel_size[i] = len(xsel[i])
        dim_ret = []
        for i in xrange(len(xsel)):
            if not N.isscalar(xsel[i]):
                if xsel[i].ndim > 0:
                    dim_ret.append(len(xsel[i]))
        ndim_ret = len(dim_ret)
        j = 0
        for i in xrange(len(xsel)):
            if not N.isscalar(xsel[i]):
                idx_shape = N.ones(ndim_ret)
                idx_shape[j] = dim_ret[i]
                xsel[i].shape = idx_shape
                j += 1

    # check if we only need basic slicing
    if not isarray and not interp:
        isbasic = True
    else:
        isbasic = False

    ret = xSelect(xsel)
    ret.isbasic = isbasic
    ret.interp = interp

    return ret            
#@-node:schmidli.20080322120238.16:numpy2xsel
#@+node:schmidli.20080322120238.17:class axisSelect
class axisSelect(object):
    """ axisSelect(inp)

    Create an axis selection object.

    Parameters:
        inp     a scalar, slice, or vector selection object

    The syntax for inp is as follows:
        [crdname|]<pre><selection><post>

    <pre> is one of:
        None    <selection> is in native coordinate space
        d       <selection> is in ISO-8601 date format
        i       <selection> is in index space

    <selection> is one of:
        #       for a scalar
        #:#:#   for a slice
        #,#,... for a vector
    where # is a number with an optional multiplier (e.g. 10k), 
    or a ISO-date. Valid multipliers include:
        k (10**3), M (10**6), h (3600), m (60), H (100)

    <post> is one of:
        i       interpolate data to exact location
        n       round to nearest index

    Examples:
        cidx = axisSelect('d20070321-09')
        cidx = axisSelect('d20070321-09:18:3h')
        cidx = axisSelect('i4')
        cidx = axisSelect('i10,20,24i')
        cidx = axisSelect('ZP|1500')
        cidx = axisSelect('20k:100k:5k')
        cidx = axisSelect('ZP|1500n')
    """
    #@  @+others
    #@+node:schmidli.20080322120238.18:__init__
    def __init__(self, inp):
        """ Overview of attributes:
            .type   one of 'scalar', 'slice', 'vector'
            .fmt    one of 'number', 'datetime'
            .iscrd  True/False
            .interp False/True
            .clip   True/False
        """

        if not isinstance(inp, str):
            raise TypeError, "Invalid argument type"
        if len(inp) == 0:
            raise ValueError, "Empty string is not a valid input"

        # default settings
        self.type = 'slice'
        self.fmt = 'number'
        self.iscrd = True
        self.clip = True

        # check prefix

        if inp[0] == 'd':
            #raise NotImplementedError, "Date/time selection is not yet implemented"
            self.fmt = 'datetime'
            inp = inp[1:]
        elif inp[0] == 'i':
            self.iscrd = False
            inp = inp[1:]

        # check postfix
        if inp[-1] in 'in':
            postfix = inp[-1]
            inp = inp[:-1]
        else:
            postfix = None

        # check if to clip field
        if inp[-1] == 'm':
            self.clip = False
            inp = inp[:-1]

        if len(inp) == 0:
            raise ValueError, "Invalid input string"   

        # check for multi-dimensional coordinate name
        inpv = inp.split('|')
        if len(inpv) > 1:
            self.mdcrd = inpv[0]
            inp = inpv[1]
            if postfix == 'n':          # default is to interpolate
                self.interp = False
            else:
                self.interp = True
        else:
            self.mdcrd = None
            inp = inpv[0]
            if postfix == 'i':          # default is to not interpolate
                self.interp = True
            else:
                self.interp = False

        # determine selection type
        inpv = inp.split(':')
        if len(inpv) > 1 and len(inpv) <= 3:
            self.type = 'slice'
        elif len(inpv) > 3:
            raise NotImplementedError, "eslice is not yet implemented"
        else:
            inpv = inp.split(',')
            if len(inpv) > 1:
                self.type = 'vector'
                if inpv[-1] == '': inpv = inpv[:-1]
            else:
                self.type = 'scalar'

        # parse the selection string
        data = []
        if self.type == 'slice' and len(inpv) == 3:
            step = inpv[2]
            if len(step) == 0:
                self.__step = None
            else:
                if step[0] == 'i':
                    self._index_step = 1
                    step = step[1:]
                self.__step = str2float(step)
                inpv = inpv[:2]
        else:
            self.__step = None
        is_first = True
        for item in inpv:
            if self.fmt == 'datetime':
                if is_first:
                    data.append(str2datetime(item))
                    is_first = False
                else:
                    data.append(str2datetime(item, templ=data[-1]))
            else:
                data.append(str2float(item))
        if self.type == 'slice':
            if len(inpv) == 1:
                data = [None, data[0]]
            elif len(inpv) == 2:
                data = [data[0], data[1]]
            if self.mdcrd is not None:
                if data[0] is None or data[1] is None or self.__step is None:
                    raise ValueError, "must specify a complete slice for multidimensional coordinates"
            if self.interp:
                if data[0] is None or data[1] is None or self.__step is None:
                    raise ValueError, "must specify a complete slice in interpolation mode"
        self.__data = data
    #@-node:schmidli.20080322120238.18:__init__
    #@+node:schmidli.20080322120238.19:__str__
    def __str__(self):
        if self.type == 'scalar':
            _str = 'scalar(' + str(self.__data[0]) + ')'
        elif self.type == 'slice':
            _str = 'slice(' + str(self.start) + ', ' + str(self.stop) + ', ' \
                            + str(self.step) + ')'
        elif self.type == 'vector':
            _str = 'vector('
            for item in self.__data:
                _str += str(item) + ' '
            _str = _str[:-1] + ')'
        _str = 'axisSelect(' + _str + ', iscrd: ' + str(self.iscrd) \
            + ', interp: ' + str(self.interp) + ')'
        return _str
    #@-node:schmidli.20080322120238.19:__str__
    #@+node:schmidli.20080322120238.20:__getitem__
    def __getitem__(self, index):
        if self.type == 'vector':
            return self.__data[index]
        else:
            raise LookupError, "Not valid method for "+self.type+"."
    #@-node:schmidli.20080322120238.20:__getitem__
    #@+node:schmidli.20080322120238.21:__len__
    def __len__(self):
        return len(self.__data)
    #@-node:schmidli.20080322120238.21:__len__
    #@+node:schmidli.20080322163604:__iter__
    def __iter__(self):
        if self.type == 'slice':
            data = N.arange(self.start, self.stop, self.step)
            if data[-1]+self.step == self.stop:
                data = N.concatenate((data, [self.stop]))
            data = data.tolist()
        elif self.type == 'scalar':
            data = (self.v)
        else:
            data = tuple(self.v)
        return iter(data)
    #@-node:schmidli.20080322163604:__iter__
    #@+node:schmidli.20080322172813:tolist
    def tolist(self):
        if self.type == 'slice':
            data = N.arange(self.start, self.stop, self.step)
            if data[-1]+self.step == self.stop:
                data = N.concatenate((data, [self.stop]))
            data = data.tolist()
        elif self.type == 'scalar':
            data = [self.v]
        else:
            data = list(self.v)
        return data
    #@-node:schmidli.20080322172813:tolist
    #@+node:schmidli.20080322120238.22:toindex
    def toindex(self, file, axis, mdcrd=None, isel=None, clip=True, ep=0.0):
        """ Convert a axisSelect object from coordinate space to index space
        """

        dimsize = None; refdate = None
        dims = None; axis_no = 0
        if self.iscrd:
            if file.cf2dims is not None:
                axfile = file.cf2dims[axis]
            else:
                axfile = axis
            if not file.file.variables.has_key(axfile):
                self.iscrd = False
        if self.iscrd:
            if mdcrd is None:
                if file.cf2dims is not None:
                    axfile = file.cf2dims[axis]
                else:
                    axfile = axis
                crd = file.file.variables[axfile]
                if self.fmt == 'datetime':
                    refdate = get_refdate(crd)
                crd = crd[:]
            else:
                crd = get_variable(file, mdcrd, isel)
                var = file.variables[mdcrd]
                dims = list(var.cf_dimensions)
                for axis2 in isel.keys():
                    if isel[axis2].type == 'scalar' and axis2 != axis: 
                        try:
                            dims.remove(axis2)
                        except ValueError, e:
                            pass
                axis_no = dims.index(axis)
                if self.type == 'scalar': dims.remove(axis)
                if var.rank < 2:
                    raise ValueError, "Coordinate variable "+self.mdcrd+ " is not multidimensional"
        else:
            dimsize = file.cf_dimensions[axis]
            crd = None
        ret = self.toindex_crd(crd, axis=axis, axis_no=axis_no, dimsize=dimsize, refdate=refdate, 
                               clip=clip, ep=0.0)
        ret.dims = dims
        ret.axis = axis
        return ret
    #@-node:schmidli.20080322120238.22:toindex
    #@+node:schmidli.20080322120238.23:toindex_crd
    def toindex_crd(self, crd, axis= None, axis_no=0, dimsize=None, refdate=None, clip=True, ep=0.0):
        """ Convert a axisSelect object from coordinate space to index space
        """

        interp = self.interp
        round_ = not interp
        clip = self.clip
        ep = 0.5

        cidx = self
        data = copy.copy(self.__data)
        idx = axisIdxSelect(self)

        # convert datetime to seconds since a reference date
        if cidx.fmt == 'datetime':
            for i in xrange(len(data)):
                if data[i] is not None:
                    data[i] = data[i] - refdate
                    data[i] = data[i].days*86400. + data[i].seconds

        # if interp=True: convert slice to vector object
        if cidx.type == 'slice' and interp:
            if data[0] is None:
                if cidx.iscrd:
                    start = crd.min()
                else:
                    if cidx.step < 0:
                        start = dimsize
                    else:
                        start = 0.0
            else:
                start = data[0]
                if start < 0: start += dimsize
            if data[1] is None: 
                if cidx.iscrd:
                    stop = crd.max()
                else:
                    if cidx.step < 0:
                        stop = 0
                    else:
                        stop = dimsize
            else:
                stop = data[1]
                if stop < 0: stop += dimsize
            if cidx.iscrd and hasattr(self,'_index_step'):
                start = _rindex(crd, start, axis=axis_no, round=round_, clip=clip, ep=ep)
                stop = _rindex(crd, stop, axis=axis_no, round=round_, clip=clip, ep=ep)
                cidx.iscrd = False

            data = N.arange(start, stop, cidx.step)
            if len(data) > 0 and data[-1]+cidx.step == stop:
                data = N.concatenate((data, [stop]))        
            idx.type = 'vector'

        # convert from coordinate to index space
        if cidx.iscrd:
            if idx.type == 'slice':
                dir = 1
                if cidx.step is not None:
                    if len(crd > 1):
                        if hasattr(self,'_index_step'):
                            if idx.step < 0: dir = -1
                            idx.step = N.round(idx.step).astype(N.int)
                        else:
                            # Note this assumes equally spaced coordinate values.
                            idx.step = cidx.step/(crd[1]-crd[0])
                            if idx.step < 0: dir = -1
                            idx.step = N.round(idx.step).astype(N.int)
                        # step size less than spacing is treated as default step (pos or neg)
                        if idx.step == 0: 
			    if dir == 1: 
                                idx.step = 1
                            else: 
                                idx.step = -1 
                    else:
                        idx.step = None
                for i in xrange(len(data)):
                    if data[i] is not None: 
                        data[i] = _rindex(crd, data[i], axis=axis_no, round=False, clip=clip, ep=ep)
                        if dir == 1:
                            if i == 0: data[i] = N.ceil(data[i]).astype(N.int)
                            if i == 1: data[i] = N.floor(data[i]).astype(N.int) + 1
                        else:
                            if i == 0: data[i] = N.floor(data[i]).astype(N.int)
                            if i == 1: 
                                data[i] = N.ceil(data[i]).astype(N.int) - 1
                                if data[i] == -1: data[i] = None
            else:
                if crd is None: raise ValueError, "Missing coordinate variable"
                if cidx.type == 'scalar':
                    data[0] = _rindex(crd, data[0], axis=axis_no, round=round_, clip=clip, ep=ep)
                else:
                    data = _rindex(crd, data, axis=axis_no, round=round_, clip=clip, ep=ep)
        else:
            if not interp:
                if idx.type == 'slice':
                    if idx.step is not None: 
                        if idx.step < 0: 
                            dir = -1
                        else:
                            dir = 1
                        idx.step = N.round(idx.step).astype(N.int)
                        if idx.step == 0: idx.step = dir
                        if idx.step < 0:
                            if data[0] is not None: 
                                data[0] = N.floor(data[0]).astype(N.int)
                                if data[0] < 0:
                                    data[0] += dimsize
                            if data[1] is not None: 
                                data[1] = N.ceil(data[1]).astype(N.int)-1
                                if data[1] < -1:
                                    data[1] += dimsize
                                elif data[1] == -1:
                                   data[1] = None
                    if idx.step is None or idx.step > 0:
                        if data[0] is not None: 
                            data[0] = N.ceil(data[0]).astype(N.int)
                            if data[0] < 0:
                                data[0] += dimsize
                        if data[1] is not None: 
                            data[1] = N.floor(data[1]).astype(N.int)+1
                            if data[1] < 0:
                                data[1] += dimsize
                else: 
                # not a slice
                    for i in xrange(len(data)):
                        data[i] = N.round(data[i]).astype(N.int)
                        if data[i] < 0: data[i] += dimsize

        if cidx.type == 'scalar' and len(data) == 1: 
            data = data[0]
            if dimsize is not None:
                # dimsize is only defined if operating in index space.
                # since later processing (xSelect.bndbox) does not handle index space values less than 0
                # convert them here.
                if data < -dimsize or data >= dimsize:    
                    raise IndexError, axis + " axis index out of range"
                elif data < 0:
                    data += dimsize
        if idx.type != 'slice': 
            if not isinstance(data, N.ma.MaskedArray):
                data = N.asarray(data)
        #if cidx.type == 'scalar' and len(data) == 1: data = data[0]
        idx.setdata(data)
        return idx
    #@-node:schmidli.20080322120238.23:toindex_crd
    #@+node:schmidli.20080322120238.24:getstart
    def getstart(self):
        """ return start value of slice object """

        if self.type == 'slice':
            return self.__data[0]
        else:
            raise AttributeError, "Not valid attribute for "+self.type+"."
    #@-node:schmidli.20080322120238.24:getstart
    #@+node:schmidli.20080322120238.25:setstart
    def setstart(self, value):
        """ return start value of slice object """

        if self.type == 'slice':
            self.__data[0] = value
        else:
            raise AttributeError, "Not valid attribute for "+self.type+"."
    #@-node:schmidli.20080322120238.25:setstart
    #@+node:schmidli.20080322120238.26:getstop
    start = property(getstart, setstart)
    def getstop(self):
        """ return stop value of slice object """

        if self.type == 'slice':
            return self.__data[1]
        else:
            raise AttributeError, "Not valid attribute for "+self.type+"."
    #@-node:schmidli.20080322120238.26:getstop
    #@+node:schmidli.20080322120238.27:setstop
    def setstop(self, value):
        """ return stop value of slice object """

        if self.type == 'slice':
            self.__data[1] = value
        else:
            raise AttributeError, "Not valid attribute for "+self.type+"."
    #@-node:schmidli.20080322120238.27:setstop
    #@+node:schmidli.20080322120238.28:getstep
    stop = property(getstop, setstop)
    def getstep(self):
        """ return step value of slice object """

        if self.type == 'slice':
            return self.__step
        else:
            raise AttributeError, "Not valid attribute for "+self.type+"."
    #@-node:schmidli.20080322120238.28:getstep
    #@+node:schmidli.20080322120238.29:setstep
    def setstep(self, value):
        """ return step value of slice object """

        if self.type == 'slice':
            self.__step = value
        else:
            raise AttributeError, "Not valid attribute for "+self.type+"."
    #@-node:schmidli.20080322120238.29:setstep
    #@+node:schmidli.20080322120238.30:getvalue
    step = property(getstep, setstep)
    def getvalue(self):
        """ return the scalar object """
        if self.type == 'scalar':
            return self.__data[0]
        elif self.type == 'slice':
            return slice(self.start, self.stop, self.step)
        elif self.type == 'vector':
            return self.__data
    #@-node:schmidli.20080322120238.30:getvalue
    #@+node:schmidli.20080322120238.31:setvalue
    def setvalue(self, value):
        """ set the scalar object """
        if isinstance(value, slice):
            self.type = 'slice'
            self.__data = [value.start, value.stop]
            self.__step = value.step
        elif N.isscalar(value):
            self.type = 'scalar'
            self.__data = [value]
        else:
            self.type = 'vector'
            self.__data = value
        #if self.type == 'scalar':
        #    self.__data[0] = value
        #elif self.type == 'vector':
        #    self.__data = value
        #else:
        #    raise AttributeError, "Not valid operation for type=slice"
    #@-node:schmidli.20080322120238.31:setvalue
    #@+node:schmidli.20080322120238.32:is_vector
    v = property(getvalue, setvalue)
    def is_vector(self):
        if self.type == 'vector':
            return True
        elif self.type == 'scalar':
            if not N.isscalar(self.__data[0]):
                return True
        return False
    #@-node:schmidli.20080322120238.32:is_vector
    #@+node:schmidli.20080322120238.33:_getdata
    isvect = property(is_vector)
    def _getdata(self):
        return copy.copy(self.__data)
    #@-node:schmidli.20080322120238.33:_getdata
    #@+node:schmidli.20080322120238.34:_getstep
    def _getstep(self):
        return self.__step
    #@-node:schmidli.20080322120238.34:_getstep
    #@-others
#@-node:schmidli.20080322120238.17:class axisSelect
#@+node:schmidli.20080322120238.35:class axisIdxSelect
class axisIdxSelect(object):
    """ axisIdxSelect(cidx)

    Create an axis selection object (index space only)

    Parameters:
        cidx    an coordinate space axis selection object (axisSelect)

    Examples:
        cidx = axisSelect('d20070321-09:18:3h')
        idx = axisIdxSelect(cidx)
    """
    #@  @+others
    #@+node:schmidli.20080322120238.36:__init__
    def __init__(self, cidx):
        """ Overview of attributes:
            .type   one of 'scalar', 'slice', 'vector'
            .interp False/True
        """

        if isinstance(cidx, axisSelect):
            # copy data members
            self.type = cidx.type
            self.interp = cidx.interp
            self.axis = None
            self.dims = None
            #self.__data = cidx._getdata()
            self.__step = cidx._getstep()
            self.__data = None
        elif isinstance(cidx, slice):
            self.type = 'slice'
            self.interp = False
            self.axis = None
            self.dims = None
            self.__data = [cidx.start, cidx.stop]
            self.__step = cidx.step
        else:
            raise TypeError, "Invalid argument type"
    #@-node:schmidli.20080322120238.36:__init__
    #@+node:schmidli.20080322120238.37:__str__
    def __str__(self):
        if self.type == 'scalar':
            _str = 'scalar(' + str(self.__data) + ')'
        elif self.type == 'slice':
            _str = 'slice(' + str(self.start) + ', ' + str(self.stop) + ', ' + str(self.step) + ')'
        elif self.type == 'vector':
            _str = 'vector('
            for item in self.__data:
                _str += str(item) + ' '
            _str = _str[:-1] + ')'
        _str = 'axisIdxSelect(' + _str + ', interp: ' + str(self.interp) \
                + ', dims: ' + str(self.dims) + ')'
        return _str
    #@-node:schmidli.20080322120238.37:__str__
    #@+node:schmidli.20080322120238.38:__getitem__
    def __getitem__(self, index):
        if self.type == 'vector':
            return self.__data[index]
        else:
            raise LookupError, "Not valid method for "+self.type+"."
    #@-node:schmidli.20080322120238.38:__getitem__
    #@+node:schmidli.20080322120238.39:getstart
    def getstart(self):
        """ return start value of slice object """

        if self.type == 'slice':
            return self.__data[0]
        else:
            raise AttributeError, "Not valid attribute for "+self.type+"."
    #@-node:schmidli.20080322120238.39:getstart
    #@+node:schmidli.20080322120238.40:setstart
    def setstart(self, value):
        """ return start value of slice object """

        if self.type == 'slice':
            self.__data[0] = value
        else:
            raise AttributeError, "Not valid attribute for "+self.type+"."
    #@-node:schmidli.20080322120238.40:setstart
    #@+node:schmidli.20080322120238.41:getstop
    start = property(getstart, setstart)
    def getstop(self):
        """ return stop value of slice object """

        if self.type == 'slice':
            return self.__data[1]
        else:
            raise AttributeError, "Not valid attribute for "+self.type+"."
    #@-node:schmidli.20080322120238.41:getstop
    #@+node:schmidli.20080322120238.42:setstop
    def setstop(self, value):
        """ return stop value of slice object """

        if self.type == 'slice':
            self.__data[1] = value
        else:
            raise AttributeError, "Not valid attribute for "+self.type+"."
    #@-node:schmidli.20080322120238.42:setstop
    #@+node:schmidli.20080322120238.43:getstep
    stop = property(getstop, setstop)
    def getstep(self):
        """ return step value of slice object """

        if self.type == 'slice':
            return self.__step
        else:
            raise AttributeError, "Not valid attribute for "+self.type+"."
    #@-node:schmidli.20080322120238.43:getstep
    #@+node:schmidli.20080322120238.44:setstep
    def setstep(self, value):
        """ return step value of slice object """

        if self.type == 'slice':
            self.__step = value
        else:
            raise AttributeError, "Not valid attribute for "+self.type+"."
    #@-node:schmidli.20080322120238.44:setstep
    #@+node:schmidli.20080322120238.45:getvalue
    step = property(getstep, setstep)
    def getvalue(self):
        """ return the scalar object """
        if self.type == 'scalar':
            return self.__data
        elif self.type == 'slice':
            return slice(self.start, self.stop, self.step)
        elif self.type == 'vector':
            return self.__data
    #@-node:schmidli.20080322120238.45:getvalue
    #@+node:schmidli.20080322120238.46:setvalue
    def setvalue(self, value):
        """ set the scalar object """
        if self.type == 'scalar':
            self.__data = value
        elif self.type == 'vector':
            self.__data = value
        else:
            raise AttributeError, "Not valid operation for type=slice"
    #@-node:schmidli.20080322120238.46:setvalue
    #@+node:schmidli.20080322120238.47:setdata
    v = property(getvalue, setvalue)
    def setdata(self, value):
        self.__data = value
    #@-node:schmidli.20080322120238.47:setdata
    #@+node:schmidli.20080322120238.48:is_array
    def is_array(self):
        if self.type == 'vector':
            return True
        elif self.type == 'scalar':
            if not N.isscalar(self.__data):
                return True
        return False
    #@-node:schmidli.20080322120238.48:is_array
    #@+node:schmidli.20080322120238.49:ndim_
    isarray = property(is_array)
    def ndim_(self):
        if not self.isarray:
            raise AttributeError, "Not valid attribute for "+self.type+"."
        return self.__data.ndim
    #@-node:schmidli.20080322120238.49:ndim_
    #@+node:schmidli.20080322120238.50:axlen_
    ndim = property(ndim_)
    def axlen_(self):
        if self.dims is not None:
            return self.dimsize(self.axis)
        else:
            if self.type == 'scalar':
                return 0
            elif self.type == 'slice':
                return -1
            else:
                return self.__data.size
    #@-node:schmidli.20080322120238.50:axlen_
    #@+node:schmidli.20080322120238.51:dimsize
    axlen = property(axlen_)
    def dimsize(self, axis):
        if self.dims is None: return 0
        if axis in self.dims:
            i = self.dims.index(axis)
            return self.__data.shape[i]
        else:
            return 0
    #@-node:schmidli.20080322120238.51:dimsize
    #@-others
#@-node:schmidli.20080322120238.35:class axisIdxSelect
#@+node:schmidli.20080322120238.52:class realSelect
class realSelect(object):
    """ realSelect(sel)

    Create an extended selection object, containing floats and
    masked arrays

    Examples:
        rsel = realSelect()[5.5, 2.2:4.4:1.1, (2.2,3,3.5),:]]
    """
    #@  @+others
    #@+node:schmidli.20080322120238.53:__init__
    def __init__(self):
        pass
    #@-node:schmidli.20080322120238.53:__init__
    #@+node:schmidli.20080322120238.54:__str__
    def __str__(self):
        _str = str(self)
        return _str
    #@-node:schmidli.20080322120238.54:__str__
    #@+node:schmidli.20080322120238.55:__getitem__
    def __getitem__(self, index):
        return index
    #@-node:schmidli.20080322120238.55:__getitem__
    #@-others
#@-node:schmidli.20080322120238.52:class realSelect
#@+node:schmidli.20080322120238.56:str2float
def str2float(inp):
    """ Convert a string with an optional postfix into a float
        postfix must be in 'kMhmH'
    """

    if inp == '': return None

    postfix = inp[-1]
    if postfix in fact_keys:
        ret = float(inp[:-1])*fact[postfix]
    else:
        ret = float(inp)
    return ret
#@-node:schmidli.20080322120238.56:str2float
#@+node:schmidli.20080322120238.57:str2datetime
def str2datetime(inp, isend=False, templ=None):
    """ Convert a iso-date string into a datetime object

        Example: '20070321T0630'
    """

    if inp == '': return None

    inpv = inp.split('-')
    date = ''; time = ''
    if len(inpv) < 2:
        time = inpv[0]
    else:
        (date, time) = inpv[:2]

    # set default values
    if templ is not None:
        d = templ
        (year, month, day) = (d.year, d.month, d.day)
        (hour, min, sec) = (0, 0, 0)
    else:
        (year, month, day) = (0, 1, 1)
        if isend:
            (hour, min, sec) = (23, 59, 59)
        else:
            (hour, min, sec) = (0, 0, 0)

    # parse date
    if len(date) == 0:
        pass
    elif len(date) <= 2:
        day = int(date)
    elif len(date) <= 4:
        month = int(date[:-2])
        day = int(date[-2:])
    elif len(date) <= 8:
        year = int(date[:-4])
        month = int(date[-4:-2])
        day = int(date[-2:])
    else:
        raise ValueError, 'Invalid date format'

    # parse time
    if len(time) == 0:
        pass
    elif len(time) <= 2:
        hour = int(time)
    elif len(time) <= 4:
        hour = int(time[:2])
        min = int(time[2:])
    elif len(time) <= 6:
        hour = int(time[:2])
        min = int(time[2:4])
        sec = int(time[4:])
    else:
        raise ValueError, 'Invalid time format'

    ret = datetime.datetime(year, month, day, hour, min, sec)
    return ret
#@-node:schmidli.20080322120238.57:str2datetime
#@+node:schmidli.20080322120238.58:get_refdate
def get_refdate(inp):
    if not hasattr(inp, 'units'):
        raise ValueError, 'Missing units attribute for time coordinate variable: ' + str(inp)
    tunits = inp.units.split(' ')
    tunits = tunits[2]+' '+tunits[3]
    ret = datetime.datetime(*time.strptime(tunits, "%Y-%m-%d %H:%M:%S")[:5])
    return ret
#@-node:schmidli.20080322120238.58:get_refdate
#@+node:schmidli.20080322120238.59:is_scalar
def is_scalar(inp):
    if isinstance(inp, N.ndarray) and inp.ndim == 0:
        return True
    elif N.isscalar(inp):
        return True
    else:
        return False
#@-node:schmidli.20080322120238.59:is_scalar
#@-others
#@-node:schmidli.20080322120238:@thin pymfio/coordsel.py
#@-leo
